# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import atexit
from functools import lru_cache
import json
import logging
import os
import platform
import uuid
import random
import sys
import time

from botocore.exceptions import ClientError, NoCredentialsError
from configparser import ConfigParser
from dataclasses import asdict, dataclass, field
from datetime import datetime
from queue import Queue, Full
from threading import Thread
from typing import Any, Callable, Dict, Optional, TypeVar, cast
from urllib import request, error

from ...job_attachments.progress_tracker import SummaryStatistics

from ._session import (
    get_monitor_id,
    get_user_and_identity_store_id,
    get_boto3_client,
    get_boto3_session,
)
from ..config import config_file
from .. import version

__cached_telemetry_client = None

logger = logging.getLogger(__name__)


# Generic function return type.
F = TypeVar("F", bound=Callable[..., Any])


def get_deadline_endpoint_url(
    config: Optional[ConfigParser] = None,
) -> str:
    # Use boto3's built-in logic to get the correct endpoint URL
    client = get_boto3_client("deadline", config=config)
    return client.meta.endpoint_url


@dataclass
class TelemetryEvent:
    """Base class for telemetry events"""

    event_type: str = "com.amazon.rum.deadline.uncategorized"
    event_details: Dict[str, Any] = field(default_factory=dict)


class TelemetryClient:
    """
    Sends telemetry events periodically to the Deadline Cloud telemetry service.

    This client holds a queue of events which is written to synchronously, and processed
    asynchronously, where events are sent in the background, so that it does not slow
    down user interactivity.

    Telemetry events contain non-personally-identifiable information that helps us
    understand how users interact with our software so we know what features our
    customers use, and/or what existing pain points are.

    Data is aggregated across a session ID (a UUID created at runtime), used to mark every
    telemetry event for the lifetime of the application), and a 'telemetry identifier' (a
    UUID recorded in the configuration file), to aggregate data across multiple application
    lifetimes on the same machine.

    Telemetry collection can be opted-out of by running:
    'deadline config set "telemetry.opt_out" true' or setting the environment variable
    'DEADLINE_CLOUD_TELEMETRY_OPT_OUT=true'
    """

    # Used for backing off requests if we encounter errors from the service.
    # See https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/
    MAX_QUEUE_SIZE = 25
    BASE_TIME = 0.5
    MAX_BACKOFF_SECONDS = 10  # The maximum amount of time to wait between retries
    MAX_RETRY_ATTEMPTS = 4

    ENDPOINT_PREFIX = "management."

    _common_details: Dict[str, Any] = {}
    _system_metadata: Dict[str, Any] = {}

    def __init__(
        self,
        package_name: str,
        package_ver: str,
        config: Optional[ConfigParser] = None,
    ):
        self._initialized: bool = False
        self.package_name = package_name
        self.package_ver = ".".join(package_ver.split(".")[:3])

        # IDs for this session
        self.session_id: str = str(uuid.uuid4())
        self.telemetry_id: str = self._get_telemetry_identifier(config=config)
        # If a different base package is provided, include info from this library as supplementary info
        if package_name != "deadline-cloud-library":
            self._common_details["deadline-cloud-version"] = version
        self._system_metadata = self._get_system_metadata(config=config)
        self.set_opt_out(config=config)
        self.initialize(config=config)

    def set_opt_out(self, config: Optional[ConfigParser] = None) -> None:
        """
        Checks whether telemetry has been opted out by checking the DEADLINE_CLOUD_TELEMETRY_OPT_OUT
        environment variable and the 'telemetry.opt_out' config file setting.
        Note the environment variable supersedes the config file setting.
        """
        env_var_value = os.environ.get("DEADLINE_CLOUD_TELEMETRY_OPT_OUT")
        if env_var_value:
            self.telemetry_opted_out = env_var_value in config_file._TRUE_VALUES
        else:
            self.telemetry_opted_out = config_file.str2bool(
                config_file.get_setting("telemetry.opt_out", config=config)
            )
        logger.info(
            "Deadline Cloud telemetry is "
            + ("not enabled." if self.telemetry_opted_out else "enabled.")
        )

    def initialize(self, config: Optional[ConfigParser] = None) -> None:
        """
        Starts up the telemetry background thread after getting settings from the boto3 client.
        Note that if this is called before boto3 is successfully configured / initialized,
        an error can be raised. In that case we silently fail and don't mark the client as
        initialized.
        """
        if self.telemetry_opted_out:
            return

        try:
            self.endpoint: str = self._get_prefixed_endpoint(
                f"{get_deadline_endpoint_url(config=config)}/2023-10-12/telemetry",
                TelemetryClient.ENDPOINT_PREFIX,
            )

            # Some environments might not have SSL, so we'll use the vendored botocore SSL context
            from botocore.httpsession import create_urllib3_context, get_cert_path

            self._urllib3_context = create_urllib3_context()
            self._urllib3_context.load_verify_locations(cafile=get_cert_path(True))

            user_id, _ = get_user_and_identity_store_id(config=config)
            if user_id:
                self._system_metadata["user_id"] = user_id

            monitor_id: Optional[str] = get_monitor_id(config=config)
            if monitor_id:
                self._system_metadata["monitor_id"] = monitor_id

            self._initialized = True
            self._start_threads()
        except Exception:
            # Silently swallow any exceptions
            return

    @property
    def is_initialized(self) -> bool:
        return self._initialized

    def _get_prefixed_endpoint(self, endpoint: str, prefix: str) -> str:
        """Insert the prefix right after 'https://'"""
        if endpoint.startswith("https://"):
            prefixed_endpoint = endpoint[:8] + prefix + endpoint[8:]
            return prefixed_endpoint
        return endpoint

    def _get_telemetry_identifier(self, config: Optional[ConfigParser] = None):
        identifier = config_file.get_setting("telemetry.identifier", config=config)
        try:
            uuid.UUID(identifier, version=4)
        except ValueError:  # Thrown if the user_id isn't in UUID4 format
            identifier = str(uuid.uuid4())
            config_file.set_setting("telemetry.identifier", identifier)
        return identifier

    def _start_threads(self) -> None:
        """Set up background threads for shutdown checking and request sending"""
        self.event_queue: Queue[Optional[TelemetryEvent]] = Queue(
            maxsize=TelemetryClient.MAX_QUEUE_SIZE
        )
        atexit.register(self._exit_cleanly)
        self.processing_thread: Thread = Thread(
            target=self._process_event_queue_thread, daemon=True
        )
        self.processing_thread.start()

    def _get_system_metadata(self, config: Optional[ConfigParser]) -> Dict[str, Any]:
        """
        Builds up a dict of non-identifiable metadata about the system environment.

        This will be used in the Rum event metadata, which has a limit of 10 unique values.
        """
        platform_info = platform.uname()
        metadata: Dict[str, Any] = {
            "service": self.package_name,
            "version": self.package_ver,
            "python_version": platform.python_version(),
            "osName": "macOS" if platform_info.system == "Darwin" else platform_info.system,
            "osVersion": platform_info.release,
        }

        return metadata

    def _exit_cleanly(self):
        self.event_queue.put(None)
        self.processing_thread.join()

    def _send_request(self, req: request.Request) -> None:
        attempts = 0
        success = False
        while not success:
            try:
                with request.urlopen(req, context=self._urllib3_context):
                    logger.debug("Successfully sent telemetry.")
                    success = True
            except error.HTTPError as httpe:
                if httpe.code == 429 or httpe.code == 500:
                    logger.debug(f"Error received from service. Waiting to retry: {str(httpe)}")

                    attempts += 1
                    if attempts >= TelemetryClient.MAX_RETRY_ATTEMPTS:
                        raise Exception("Max retries reached sending telemetry")

                    backoff_sleep = random.uniform(
                        0,
                        min(
                            TelemetryClient.MAX_BACKOFF_SECONDS,
                            TelemetryClient.BASE_TIME * 2**attempts,
                        ),
                    )
                    time.sleep(backoff_sleep)
                else:  # Reraise any exceptions we didn't expect
                    raise

    def _process_event_queue_thread(self):
        """Background thread for processing the telemetry event data queue and sending telemetry requests."""
        while True:
            # Blocks until we get a new entry in the queue
            event_data: Optional[TelemetryEvent] = self.event_queue.get()
            # We've received the shutdown signal
            if event_data is None:
                return

            headers = {"Accept": "application-json", "Content-Type": "application-json"}
            try:
                request_body = {
                    "BatchId": str(uuid.uuid4()),
                    "RumEvents": [
                        {
                            "details": str(json.dumps(event_data.event_details)),
                            "id": str(uuid.uuid4()),
                            "metadata": str(json.dumps(self._system_metadata)),
                            "timestamp": int(datetime.now().timestamp()),
                            "type": event_data.event_type,
                        },
                    ],
                    "UserDetails": {"sessionId": self.session_id, "userId": self.telemetry_id},
                }
                request_body_encoded = str(json.dumps(request_body)).encode("utf-8")
            except Exception as exc:
                logger.debug(f"Failed to serialize telemetry data. {str(exc)}")
                continue

            req = request.Request(url=self.endpoint, data=request_body_encoded, headers=headers)
            try:
                logger.debug("Sending telemetry data: %s", request_body)
                self._send_request(req)
            except Exception as exc:
                # Swallow any kind of uncaught exception and stop sending telemetry
                logger.debug(f"Error received from service. {str(exc)}")
                return
            self.event_queue.task_done()

    def _put_telemetry_record(self, event: TelemetryEvent) -> None:
        if not self._initialized or self.telemetry_opted_out:
            return
        try:
            self.event_queue.put_nowait(event)
        except Full:
            # Silently swallow the error if the event queue is full (due to throttling of the service)
            pass

    def _record_summary_statistics(
        self, event_type: str, summary: SummaryStatistics, from_gui: bool
    ):
        details: Dict[str, Any] = asdict(summary)
        self.record_event(event_type=event_type, event_details=details, from_gui=from_gui)

    def record_hashing_summary(self, summary: SummaryStatistics, *, from_gui: bool = False):
        self._record_summary_statistics(
            "com.amazon.rum.deadline.job_attachments.hashing_summary", summary, from_gui
        )

    def record_upload_summary(self, summary: SummaryStatistics, *, from_gui: bool = False):
        self._record_summary_statistics(
            "com.amazon.rum.deadline.job_attachments.upload_summary", summary, from_gui
        )

    def record_error(
        self, event_details: Dict[str, Any], exception_type: str, from_gui: bool = False
    ):
        event_details["exception_type"] = exception_type
        # Possibility to add stack trace here
        self.record_event("com.amazon.rum.deadline.error", event_details, from_gui=from_gui)

    def record_event(
        self, event_type: str, event_details: Dict[str, Any], *, from_gui: bool = False
    ):
        try:
            self.update_common_details({"accountId": self.get_account_id(get_boto3_session())})
        except Exception as e:
            # Print any errors when getting the boto3 session, then proceed
            logger.debug(f"Could not add account ID to telemetry: {str(e)}")

        event_details.update(self._common_details)
        event_details["usage_mode"] = "GUI" if from_gui else "CLI"
        self._put_telemetry_record(
            TelemetryEvent(
                event_type=event_type,
                event_details=event_details,
            )
        )

    @lru_cache
    def get_account_id(self, boto3_session) -> Optional[str]:
        """
        Retrieves the AWS account ID for the current user.

        If the user is not authenticated, print an error message
        """
        try:
            return boto3_session.client("sts").get_caller_identity()["Account"]
        except (ClientError, NoCredentialsError) as e:
            print(f"Could not add account ID to telemetry: {str(e)}")
            return None

    def update_common_details(self, details: Dict[str, Any]):
        """Updates the dict of common data that is included in every telemetry request."""
        self._common_details.update(details)


def get_telemetry_client(
    package_name: str, package_ver: str, config: Optional[ConfigParser] = None
) -> TelemetryClient:
    """
    Retrieves the cached telemetry client, lazy-loading the first time this is called.
    :param package_name: Base package name to associate data by.
    :param package_ver: Base package version to associate data by.
    :param config: Optional configuration to use for the client. Loads defaults if not given.
    :return: Telemetry client to make requests with.
    """
    global __cached_telemetry_client
    if not __cached_telemetry_client:
        __cached_telemetry_client = TelemetryClient(
            package_name=package_name,
            package_ver=package_ver,
            config=config,
        )
    elif not __cached_telemetry_client.is_initialized:
        __cached_telemetry_client.initialize(config=config)

    return __cached_telemetry_client


def get_deadline_cloud_library_telemetry_client(
    config: Optional[ConfigParser] = None,
) -> TelemetryClient:
    """
    Retrieves the cached telemetry client, specifying the Deadline Cloud Client Library's package information.
    :param config: Optional configuration to use for the client. Loads defaults if not given.
    :return: Telemetry client to make requests with.
    """
    return get_telemetry_client("deadline-cloud-library", version, config=config)


def record_success_fail_telemetry_event(**decorator_kwargs: Any) -> Callable[[F], F]:
    """
    Decorator to try catch a function. Sends a success / fail telemetry event.
    :param ** Python variable arguments. See https://docs.python.org/3/glossary.html#term-parameter.
    """

    def inner(function: F) -> F:
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            """
            Wrapper to try-catch a function for telemetry
            :param * Python variable argument. See https://docs.python.org/3/glossary.html#term-parameter
            :param ** Python variable argument. See https://docs.python.org/3/glossary.html#term-parameter
            """
            success: bool = False
            try:
                result = function(*args, **kwargs)
                success = True
                return result
            finally:
                event_name = decorator_kwargs.get("metric_name", function.__name__)

                event_details: dict = decorator_kwargs.get("event_details", {})
                event_details["is_success"] = success
                raised_exception = sys.exc_info()[1]
                if raised_exception is not None:
                    event_details["exception_type"] = type(raised_exception).__name__

                get_deadline_cloud_library_telemetry_client().record_event(
                    event_type=f"com.amazon.rum.deadline.{event_name}",
                    event_details=event_details,
                )

        wrapper.__doc__ = function.__doc__
        return cast(F, wrapper)

    return inner


def record_function_latency_telemetry_event(**decorator_kwargs: Any) -> Callable[[F], F]:
    """
    Decorator to time a function. Sends a latency telemetry event.
    :param ** Python variable arguments. See https://docs.python.org/3/glossary.html#term-parameter.
    """

    def inner(function: F) -> F:
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            """
            Wrapper to time a function for latency telemetry
            :param * Python variable argument. See https://docs.python.org/3/glossary.html#term-parameter
            :param ** Python variable argument. See https://docs.python.org/3/glossary.html#term-parameter
            """
            start_t = time.perf_counter_ns()
            ret_val = function(*args, **kwargs)
            end_t = time.perf_counter_ns()

            latency = end_t - start_t

            event_name = decorator_kwargs.get("metric_name", function.__name__)
            get_deadline_cloud_library_telemetry_client().record_event(
                event_type="com.amazon.rum.deadline.latency",
                event_details={"latency": latency, "function_call": event_name},
            )

            return ret_val

        wrapper.__doc__ = function.__doc__
        return cast(F, wrapper)

    return inner



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Runs the Deadline CLI. Can be run as a python script file.
Required by pyinstaller, do not delete.
"""


def main() -> None:
    from deadline.client.cli._deadline_cli import main

    main(
        # Override the program name to always be "deadline"
        prog_name="deadline",
    )


if __name__ == "__main__":
    main()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""Functions for handling and retrieving AWS clients."""

from __future__ import annotations

from functools import lru_cache
from typing import Optional

import boto3
import botocore
from boto3.s3.transfer import create_transfer_manager
from botocore.client import BaseClient, Config

from deadline.client.config import config_file

from .. import version
from ..exceptions import AssetSyncError
from .aws_config import (
    S3_CONNECT_TIMEOUT_IN_SECS,
    S3_READ_TIMEOUT_IN_SECS,
    S3_RETRIES_MODE,
    VENDOR_CODE,
)

MAX_SIZE_CACHE = 128


# Should create a new botocore session since botocore session may be modified by boto3 session/client using it
# https://github.com/boto/boto3/blob/61de529b5f9a7bdcc8c76debb472a7f934d048e6/boto3/session.py#L79
def get_botocore_session() -> botocore.session.Session:
    return botocore.session.get_session()


@lru_cache(maxsize=MAX_SIZE_CACHE)
def get_boto3_session(
    botocore_session: botocore.session.Session = get_botocore_session(),
) -> boto3.session.Session:
    return boto3.session.Session(botocore_session=botocore_session)


@lru_cache(maxsize=MAX_SIZE_CACHE)
def get_deadline_client(
    session: Optional[boto3.session.Session] = None, endpoint_url: Optional[str] = None
) -> BaseClient:
    """
    Get a boto3 Deadline client to make API calls to Deadline
    """
    if session is None:
        session = get_boto3_session()

    return session.client(VENDOR_CODE, endpoint_url=endpoint_url)


@lru_cache(maxsize=MAX_SIZE_CACHE)
def get_s3_client(session: Optional[boto3.Session] = None) -> BaseClient:
    """
    Get a boto3 S3 client to make API calls to S3
    """
    if session is None:
        session = get_boto3_session()

    s3_max_pool_connections = get_s3_max_pool_connections()

    client = session.client(
        "s3",
        config=Config(
            signature_version="s3v4",
            connect_timeout=S3_CONNECT_TIMEOUT_IN_SECS,
            read_timeout=S3_READ_TIMEOUT_IN_SECS,
            retries={"mode": S3_RETRIES_MODE},
            user_agent_extra=f"S3A/Deadline/NA/JobAttachments/{version}",
            max_pool_connections=s3_max_pool_connections,
        ),
        endpoint_url=f"https://s3.{session.region_name}.amazonaws.com",
    )

    def add_expected_bucket_owner(params, model, **kwargs):
        """
        Add the expected bucket owner to the params if the API operation to run can use it.
        """
        if "ExpectedBucketOwner" in model.input_shape.members:
            params["ExpectedBucketOwner"] = get_account_id(session=session)

    client.meta.events.register("provide-client-params.s3.*", add_expected_bucket_owner)

    return client


def get_s3_max_pool_connections() -> int:
    try:
        s3_max_pool_connections = int(config_file.get_setting("settings.s3_max_pool_connections"))
    except ValueError as ve:
        raise AssetSyncError(
            "Failed to parse configuration settings. Please ensure that the following settings in the config file are integers: "
            "'s3_max_pool_connections'"
        ) from ve
    if s3_max_pool_connections <= 0:
        raise AssetSyncError(
            f"Nonvalid value for configuration setting: 's3_max_pool_connections' ({s3_max_pool_connections}) must be positive integer."
        )
    return s3_max_pool_connections


@lru_cache(maxsize=MAX_SIZE_CACHE)
def get_s3_transfer_manager(s3_client: BaseClient):
    transfer_config = boto3.s3.transfer.TransferConfig()
    return create_transfer_manager(client=s3_client, config=transfer_config)


@lru_cache(maxsize=MAX_SIZE_CACHE)
def get_sts_client(session: Optional[boto3.session.Session] = None) -> BaseClient:
    """
    Get a boto3 sts client to make API calls to STS.
    """
    if session is None:
        session = get_boto3_session()
    return session.client(
        "sts",
        endpoint_url=f"https://sts.{session.region_name}.amazonaws.com",
    )


@lru_cache(maxsize=MAX_SIZE_CACHE)
def get_caller_identity(session: Optional[boto3.session.Session] = None) -> dict[str, str]:
    """
    Get the caller identity for the current session.
    """
    return get_sts_client(session).get_caller_identity()


def get_account_id(session: Optional[boto3.session.Session] = None) -> str:
    """
    Get the account id for the current session.
    """
    return get_caller_identity(session)["Account"]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
from __future__ import annotations
import copy
import os
from logging import getLogger
from typing import Any, Optional, Dict

from qtpy.QtCore import Qt  # pylint: disable=import-error
from qtpy.QtWidgets import (  # pylint: disable=import-error; type: ignore
    QApplication,
    QFileDialog,
    QMainWindow,
)

from ..exceptions import DeadlineOperationError
from ..job_bundle.loader import (
    parse_yaml_or_json_content,
    read_yaml_or_json,
    read_yaml_or_json_object,
    validate_directory_symlink_containment,
)
from ..job_bundle.saver import save_yaml_or_json_to_file
from ..job_bundle.parameters import (
    JobParameter,
    apply_job_parameters,
    merge_queue_job_parameters,
    read_job_bundle_parameters,
)
from .dataclasses import JobBundleSettings
from .dialogs.submit_job_to_deadline_dialog import (
    SubmitJobToDeadlineDialog,
    JobBundlePurpose,
)
from .widgets.job_bundle_settings_tab import JobBundleSettingsWidget
from ..job_bundle.submission import AssetReferences
from ..api._session import session_context

logger = getLogger(__name__)


def show_job_bundle_submitter(
    *,
    input_job_bundle_dir: str = "",
    browse: bool = False,
    parent=None,
    f=Qt.WindowFlags(),
    submitter_name: Optional[str] = None,
    known_asset_paths: Optional[list[str]] = None,
) -> Optional[SubmitJobToDeadlineDialog]:
    """
    Opens an AWS Deadline Cloud job submission dialog for the provided job bundle.

    Pass f=Qt.Tool if running it within an application context and want it
    to stay on top.
    """

    if not submitter_name:
        submitter_name = "JobBundle"

    session_context["submitter-name"] = submitter_name

    if parent is None:
        # Get the main application window so we can parent ours to it
        app = QApplication.instance()
        main_windows = [
            widget
            for widget in app.topLevelWidgets()
            if isinstance(widget, QMainWindow)  # type: ignore[union-attr]
        ]
        if main_windows:
            parent = main_windows[0]

    if not input_job_bundle_dir:
        input_job_bundle_dir = QFileDialog.getExistingDirectory(
            parent, "Choose job bundle directory", input_job_bundle_dir
        )
        if not input_job_bundle_dir:
            return None

    def on_create_job_bundle_callback(
        widget: SubmitJobToDeadlineDialog,
        job_bundle_dir: str,
        settings: JobBundleSettings,
        queue_parameters: list[JobParameter],
        asset_references: AssetReferences,
        host_requirements: Optional[Dict[str, Any]] = None,
        purpose: JobBundlePurpose = JobBundlePurpose.SUBMISSION,
    ) -> dict[str, Any]:
        """
        Perform a submission when the submit button is pressed

        Args:
            widget (SubmitJobToDeadlineDialog): The Deadline job submission dialog.
            settings (JobBundleSettings): A settings object that was populated from the job submission dialog.
            job_bundle_dir (str): The directory within which to create the job bundle.
            asset_references (FlatAssetReferences): The input from the attachments provided during
                construction and the user's input in the Job Attachments tab.
        """
        # Copy the template
        file_contents, file_type = read_yaml_or_json(
            settings.input_job_bundle_dir, "template", True
        )

        template = parse_yaml_or_json_content(
            file_contents, file_type, settings.input_job_bundle_dir, "template"
        )
        template["name"] = settings.name
        if settings.description:
            template["description"] = settings.description
        else:
            # remove description field since it can't be empty
            # ignore if description is missing from template
            template.pop("description", None)

        # If "HostRequirements" is provided, inject it into each of the "Step"
        if host_requirements:
            # for each step in the template, append the same host requirements.
            for step in template["steps"]:
                step["hostRequirements"] = copy.deepcopy(host_requirements)

        # First filter the queue parameters to exclude any from the job template,
        # then extend it with the job template parameters.
        job_parameter_names = {param["name"] for param in settings.parameters}
        parameter_values: list[dict[str, Any]] = [
            {"name": param["name"], "value": param["value"]}
            for param in queue_parameters
            if param["name"] not in job_parameter_names
        ]
        parameter_values.extend(
            {"name": param["name"], "value": param["value"]} for param in settings.parameters
        )

        parameters = merge_queue_job_parameters(
            queue_parameters=queue_parameters,
            job_parameters=settings.parameters,
        )

        apply_job_parameters(
            parameter_values,
            job_bundle_dir,
            parameters,
            AssetReferences(),
        )

        save_yaml_or_json_to_file(
            bundle_dir=job_bundle_dir, filename="template", file_type=file_type, data=template
        )
        save_yaml_or_json_to_file(
            bundle_dir=job_bundle_dir,
            filename="asset_references",
            file_type=file_type,
            data=asset_references.to_dict(),
        )

        return {
            "known_asset_paths": [os.path.abspath(settings.input_job_bundle_dir)],
            "job_parameters": parameter_values,
        }

    # Ensure the job bundle doesn't contain files that resolve outside of the bundle directory
    validate_directory_symlink_containment(input_job_bundle_dir)

    # Load the template to get the starting name
    template = read_yaml_or_json_object(input_job_bundle_dir, "template", True)

    asset_references_obj = (
        read_yaml_or_json_object(input_job_bundle_dir, "asset_references", False) or {}
    )
    asset_references = AssetReferences.from_dict(asset_references_obj)

    name = "Job bundle submission"
    if template:
        name = template.get("name", name)

    if not os.path.isdir(input_job_bundle_dir):
        raise DeadlineOperationError(f"Input Job Bundle Dir is not valid: {input_job_bundle_dir}")
    initial_settings = JobBundleSettings(input_job_bundle_dir=input_job_bundle_dir, name=name)
    initial_settings.parameters = read_job_bundle_parameters(input_job_bundle_dir)
    initial_settings.browse_enabled = browse

    # Populate the initial queue parameter values based on the job template parameter values
    initial_shared_parameter_values = {}
    for parameter in initial_settings.parameters:
        if "default" in parameter or "value" in parameter:
            initial_shared_parameter_values[parameter["name"]] = parameter.get(
                "value", parameter.get("default")
            )

    submitter_dialog = SubmitJobToDeadlineDialog(
        job_setup_widget_type=JobBundleSettingsWidget,
        initial_job_settings=initial_settings,
        show_host_requirements_tab=True,
        initial_shared_parameter_values=initial_shared_parameter_values,
        auto_detected_attachments=asset_references,
        attachments=AssetReferences(),
        on_create_job_bundle_callback=on_create_job_bundle_callback,
        parent=parent,
        f=f,
        submitter_name=submitter_name,
        known_asset_paths=known_asset_paths,
    )
    submitter_dialog.show()
    return submitter_dialog



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
from __future__ import annotations

import re


def parse_frame_range(frame_string: str) -> list[int]:
    framelist_re = re.compile(r"^(?P<start>-?\d+)(-(?P<stop>-?\d+)(:(?P<step>-?\d+))?)?$")
    match = framelist_re.match(frame_string)
    if not match:
        raise ValueError("Framelist not valid")

    start = int(match.group("start"))
    stop = int(match.group("stop")) if match.group("stop") is not None else start
    frame_step = (
        int(match.group("step")) if match.group("step") is not None else 1 if start < stop else -1
    )

    if frame_step == 0:
        raise ValueError("Frame step cannot be zero")
    if start > stop and frame_step > 0:
        raise ValueError("Start frame must be less than stop frame if step is positive")
    if start < stop and frame_step < 0:
        raise ValueError("Start frame must be greater than stop frame if step is negative")

    return list(range(start, stop + (1 if frame_step > 0 else -1), frame_step))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
This module encapsulate the configuration of AWS Deadline Cloud on a workstation.

By default, configuration is stored in `~/.deadline/config`. If a user sets
the environment variable DEADLINE_CONFIG_FILE_PATH, it is used as the configuration
file path instead.
"""

__all__ = [
    "get_setting_default",
    "get_setting",
    "set_setting",
    "clear_setting",
    "get_best_profile_for_farm",
    "str2bool",
    "DEFAULT_DEADLINE_ENDPOINT_URL",
]

from .config_file import (
    DEFAULT_DEADLINE_ENDPOINT_URL,
    get_best_profile_for_farm,
    get_setting,
    get_setting_default,
    set_setting,
    clear_setting,
    str2bool,
)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Helper functions to enable submission of a Job Bundle to CreateJob
"""

from __future__ import annotations
import dataclasses
import logging
import os
from typing import Any, Tuple, Optional

from ..exceptions import DeadlineOperationError
from .parameters import JobParameter

logger = logging.getLogger(__name__)

DEFAULT_APP_NAME = "deadline"
DEFAULT_SUPPORTED_APP_PARAMETER_NAMES = [
    "targetTaskRunStatus",
    "priority",
    "maxFailedTasksCount",
    "maxRetriesPerTask",
    "maxWorkerCount",
]


@dataclasses.dataclass
class AssetReferences:
    """Holds the asset references for a job bundle."""

    input_filenames: set[str] = dataclasses.field(default_factory=set)
    """Filenames whose file contents are input to the job."""
    input_directories: set[str] = dataclasses.field(default_factory=set)
    """Directories whose contents are input to the job."""
    output_directories: set[str] = dataclasses.field(default_factory=set)
    """Directories whose contents are output from the job."""
    referenced_paths: set[str] = dataclasses.field(default_factory=set)
    """Paths that are referenced by the job, but not necessarily input or output."""

    def __init__(
        self,
        *,
        input_filenames: Optional[set[str]] = None,
        input_directories: Optional[set[str]] = None,
        output_directories: Optional[set[str]] = None,
        referenced_paths: Optional[set[str]] = None,
    ):
        self.input_filenames = input_filenames or set()
        self.input_directories = input_directories or set()
        self.output_directories = output_directories or set()
        self.referenced_paths = referenced_paths or set()

    def __bool__(self) -> bool:
        """Returns whether the object has any asset references."""
        return (
            bool(self.input_filenames)
            or bool(self.input_directories)
            or bool(self.output_directories)
            or bool(self.referenced_paths)
        )

    def union(self, other: AssetReferences):
        """Returns the union of the asset references."""
        return AssetReferences(
            input_filenames=self.input_filenames.union(other.input_filenames),
            input_directories=self.input_directories.union(other.input_directories),
            output_directories=self.output_directories.union(other.output_directories),
            referenced_paths=self.referenced_paths.union(other.referenced_paths),
        )

    @classmethod
    def from_dict(cls, obj: Optional[dict[str, Any]]) -> AssetReferences:
        if obj:
            input_filenames = obj["assetReferences"].get("inputs", {}).get("filenames", [])
            input_directories = obj["assetReferences"].get("inputs", {}).get("directories", [])
            output_directories = obj["assetReferences"].get("outputs", {}).get("directories", [])
            referenced_paths = obj["assetReferences"].get("referencedPaths", [])

            return cls(
                input_filenames=set(os.path.normpath(path) for path in input_filenames),
                input_directories=set(os.path.normpath(path) for path in input_directories),
                output_directories=set(os.path.normpath(path) for path in output_directories),
                referenced_paths=set(os.path.normpath(path) for path in referenced_paths),
            )
        else:
            return cls()

    def to_dict(self) -> dict[str, Any]:
        return {
            "assetReferences": {
                "inputs": {
                    "directories": sorted(self.input_directories),
                    "filenames": sorted(self.input_filenames),
                },
                "outputs": {"directories": sorted(self.output_directories)},
                "referencedPaths": sorted(self.referenced_paths),
            }
        }


def split_parameter_args(
    parameters: list[JobParameter],
    job_bundle_dir: str,
    app_name: Optional[str] = None,
    supported_app_parameter_names: Optional[list[str]] = None,
) -> Tuple[dict[str, Any], dict[str, Any]]:
    """
    Splits the input job_bundle_parameters into separate application paramters
    and job specific parameters.

    Args:
        parameters (list): The list of submission parameters.
        job_bundle_dir (str): The job bundle directory, used for error messages.
        app_name (str): The name of the application prefix to accept for application-specific
            parameters.
        supported_app_parameter_names (list[str]): The list of parameter names
            that can be provided with the `app_name` prefix.
    """

    if app_name is None:
        app_name = DEFAULT_APP_NAME
    if supported_app_parameter_names is None:
        supported_app_parameter_names = DEFAULT_SUPPORTED_APP_PARAMETER_NAMES

    job_parameters: dict[str, Any] = {}
    app_parameters: dict[str, Any] = {}

    if parameters:
        for parameter in parameters:
            if "value" in parameter:
                parameter_name = parameter["name"]
                parameter_value = parameter["value"]

                if parameter_name.startswith(f"{app_name}:"):
                    # Application-specific parameters
                    app_parameter_name = parameter_name.split(":", 1)[1]
                    if app_parameter_name in supported_app_parameter_names:
                        app_parameters[app_parameter_name] = parameter_value
                    else:
                        raise DeadlineOperationError(
                            f"Unrecognized parameter named {parameter_name!r} from job bundle:\n{job_bundle_dir}"
                        )
                elif ":" in parameter_name:
                    # Drop application-specific parameters from other applications
                    pass
                else:
                    parameter_type = parameter["type"].lower()
                    job_parameters[parameter_name] = {parameter_type: str(parameter_value)}

    return app_parameters, job_parameters



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Provides an object that can be used to track current status of AWS Deadline Cloud
authentication.

The object emits the following Qt Signals:
   aws_creds_changed: The AWS credentials in ~/.aws changed.
   deadline_config_changed: The AWS Deadline Cloud configuration in ~/.deadline changed.
   creds_source_changed: triggered when credential source changes
   auth_status_changed: triggered when authentication status changes
   api_availability_changed: triggered when api availability changes

The status includes three parts:
  1. Are credentials configured and available for use?
     This is checked with an sts:GetCallerIdentity AWS API call.
  2. Do the credentials grant access to AWS Deadline Cloud APIs?
     This is checked with a simplified deadline:ListFarms AWS API call.
  3. Do the credentials use Deadline Cloud monitor?
     This is checked by looking for the relevant properties
     in the AWS profile configuration.
"""

import os
import threading
from configparser import ConfigParser
from logging import getLogger
from typing import Optional

from qtpy.QtCore import QObject, QFileSystemWatcher, Signal

from .. import api
from ..config import config_file

logger = getLogger(__name__)

_deadline_authentication_status = None


class DeadlineAuthenticationStatus(QObject):
    """
    Holds status information about AWS Deadline Cloud credentials.
    Currently status values are available as properties:

       status.creds_source: result of api.get_credentials_source()
       status.auth_status: result of api.check_authentication_status()
       status.api_availability: result of api.check_deadline_api_available()

    To initialize the status of a non-default AWS Deadline Cloud configuration, pass in
    an AWS Deadline Cloud configuration object to config, call set_config to change it.
    """

    # This signal is sent when an AWS credential changes (e.g. config file)
    aws_creds_changed = Signal()
    # This signal is sent when the AWS Deadline Cloud configuration changes
    deadline_config_changed = Signal()

    # This signal is sent when an AWS authentication type changes
    creds_source_changed = Signal()
    # This signal is sent when an AWS authentication status changes
    auth_status_changed = Signal()
    # This signal is sent when AWS Deadline Cloud API availability changes
    api_availability_changed = Signal()

    @staticmethod
    def getInstance():
        global _deadline_authentication_status
        if _deadline_authentication_status is None:
            _deadline_authentication_status = DeadlineAuthenticationStatus()
        return _deadline_authentication_status

    def __init__(self, parent=None) -> None:
        super(DeadlineAuthenticationStatus, self).__init__(parent)

        self.__creds_source: Optional[api.AwsCredentialsSource] = None
        self.__auth_status: Optional[api.AwsAuthenticationStatus] = None
        self.__api_availability: Optional[bool] = None

        # Load the default config
        self.config = ConfigParser()
        self.config.read_dict(config_file.read_config())

        # Watch the ~/.aws path for any changes to config or credentials, and
        # the ~/.deadline path for any changes to the AWS Deadline Cloud config.
        self.aws_creds_file_watcher = QFileSystemWatcher()
        self.aws_creds_paths = [
            os.path.expanduser(os.path.join("~", ".aws")),
        ]
        self.deadline_config_paths = [
            os.path.expanduser(os.path.join("~", ".deadline")),
        ]
        failed_paths = self.aws_creds_file_watcher.addPaths(
            self.aws_creds_paths + self.deadline_config_paths
        )
        if failed_paths:
            logger.error(
                "Failed to watch these AWS Deadline Cloud configurations: %s", failed_paths
            )
        self.aws_creds_file_watcher.fileChanged.connect(self.files_changed)
        self.aws_creds_file_watcher.directoryChanged.connect(self.files_changed)

        self.refresh_status()

    def set_config(self, config: Optional[ConfigParser]) -> None:
        """
        Changes the AWS Deadline Cloud configuration object used to display authentication
        status.

        Args:
            config (ConfigParser): The AWS Deadline Cloud configuration to use.
        """

        # Refresh the status if any setting that impacts authentication was changed
        if self.config:
            auth_config_changed = False
            for setting_name in [
                "defaults.aws_profile_name",
            ]:
                if config_file.get_setting(setting_name, self.config) != config_file.get_setting(
                    setting_name, config
                ):
                    auth_config_changed = True
        else:
            auth_config_changed = True

        # Make a copy of the config object
        self.config = ConfigParser()
        if config:
            self.config.read_dict(config)
        else:
            self.config.read_dict(config_file.read_config())

        if auth_config_changed:
            self.refresh_status()

    @property
    def creds_source(self) -> Optional[api.AwsCredentialsSource]:
        return self.__creds_source

    @property
    def auth_status(self) -> Optional[api.AwsAuthenticationStatus]:
        return self.__auth_status

    @property
    def api_availability(self) -> Optional[bool]:
        return self.__api_availability

    def files_changed(self, changed_path) -> None:
        # Force the cached boto3 session to refresh, since we don't check the creds
        # file
        if changed_path in self.aws_creds_paths:
            logger.info(f"Path {changed_path} changed, refreshing authentication status")
            # Send it to another thread to avoid blocking the Qt event loop
            self.session_thread = threading.Thread(
                target=self._get_session, kwargs={"changed_path": changed_path}
            )
            self.session_thread.start()
        else:
            logger.info(f"Path {changed_path} changed, does not affect authentication status")

    def _get_session(self, changed_path):
        api.get_boto3_session(force_refresh=True)
        self.refresh_status()

        if changed_path in self.aws_creds_paths:
            self.aws_creds_changed.emit()
        elif changed_path in self.deadline_config_paths:
            self.deadline_config_changed.emit()

    def _refresh_creds_source(self) -> None:
        self.__creds_source = None
        self.creds_source_changed.emit()
        self.__creds_source = api.get_credentials_source(config=self.config)
        self.creds_source_changed.emit()

    def _refresh_auth_status(self) -> None:
        self.__auth_status = None
        self.auth_status_changed.emit()
        try:
            self.__auth_status = api.check_authentication_status(config=self.config)
        except BaseException as e:
            logger.exception(e)
            self.__auth_status = api.AwsAuthenticationStatus.CONFIGURATION_ERROR
        self.auth_status_changed.emit()

    def _refresh_api_availability(self) -> None:
        self.__api_availability = None
        self.api_availability_changed.emit()
        try:
            self.__api_availability = api.check_deadline_api_available(config=self.config)
        except BaseException as e:
            logger.exception(e)
            self.__api_availability = False
        self.api_availability_changed.emit()

    def refresh_status(self) -> None:
        """
        Initiates an asynchronous status refresh.
        """
        self.__creds_source_thread = threading.Thread(target=self._refresh_creds_source)
        self.__creds_source_thread.start()
        self.__auth_status_thread = threading.Thread(target=self._refresh_auth_status)
        self.__auth_status_thread.start()
        self.__api_availability_thread = threading.Thread(target=self._refresh_api_availability)
        self.__api_availability_thread.start()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import click
import json
import typing as t


class ClickLogger:
    """
    Wrapper around click that is JSON aware. Users can instantiate this as a
    replacement for using `click.echo`. A helper JSON function is also provided
    to output JSON.
    """

    def __init__(self, is_json: bool):
        self._is_json = is_json

    def is_json(self) -> bool:
        """
        Is logging in JSON mode.
        """
        return self._is_json

    def echo(
        self,
        message: t.Optional[t.Any] = None,
        file: t.Optional[t.IO[t.Any]] = None,
        nl: bool = True,
        err: bool = False,
        color: t.Optional[bool] = None,
    ):
        if not self._is_json:
            click.echo(message, file, nl, err, color)

    def json(
        self,
        message: t.Optional[dict] = None,
        file: t.Optional[t.IO[t.Any]] = None,
        nl: bool = True,
        err: bool = False,
        color: t.Optional[bool] = None,
        indent=None,
    ):
        if self._is_json:
            click.echo(json.dumps(obj=message, indent=indent), file, nl, err, color)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""Module that defines the v2023-03-03 version of the asset manifest"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Type

from .._canonical_json import canonical_path_comparator, manifest_to_canonical_json_string
from ..base_manifest import BaseAssetManifest, BaseManifestPath
from ..hash_algorithms import HashAlgorithm
from ..manifest_model import BaseManifestModel
from ..versions import ManifestVersion
from ...exceptions import ManifestDecodeValidationError


SUPPORTED_HASH_ALGS: set[HashAlgorithm] = {HashAlgorithm.XXH128}
DEFAULT_HASH_ALG: HashAlgorithm = HashAlgorithm.XXH128


@dataclass
class ManifestPath(BaseManifestPath):
    """
    Extension for version v2023-03-03 of the asset manifest.
    """

    manifest_version = ManifestVersion.v2023_03_03

    def __init__(self, *, path: str, hash: str, size: int, mtime: int) -> None:
        super().__init__(path=path, hash=hash, size=size, mtime=mtime)


@dataclass
class AssetManifest(BaseAssetManifest):
    """Version v2023-03-03 of the asset manifest"""

    totalSize: int  # pyline: disable=invalid-name

    def __init__(
        self, *, hash_alg: HashAlgorithm, paths: list[BaseManifestPath], total_size: int
    ) -> None:
        if hash_alg not in SUPPORTED_HASH_ALGS:
            raise ManifestDecodeValidationError(
                f"Unsupported hashing algorithm: {hash_alg}. Must be one of: {[e.value for e in SUPPORTED_HASH_ALGS]}"
            )

        super().__init__(hash_alg=hash_alg, paths=paths)
        self.totalSize = total_size
        self.manifestVersion = ManifestVersion.v2023_03_03

    @classmethod
    def decode(cls, *, manifest_data: dict[str, Any]) -> AssetManifest:
        """
        Return an instance of this class given a manifest dictionary.
        Assumes the manifest has been validated prior to calling.
        """
        try:
            hash_alg: HashAlgorithm = HashAlgorithm(manifest_data["hashAlg"])
        except ValueError:
            raise ManifestDecodeValidationError(
                f"Unsupported hashing algorithm: {hash_alg}. Must be one of: {[e.value for e in SUPPORTED_HASH_ALGS]}"
            )

        return cls(
            hash_alg=hash_alg,
            paths=[
                ManifestPath(
                    path=path["path"], hash=path["hash"], size=path["size"], mtime=path["mtime"]
                )
                for path in manifest_data["paths"]
            ],
            total_size=manifest_data["totalSize"],
        )

    @classmethod
    def get_default_hash_alg(cls) -> HashAlgorithm:  # pragma: no cover
        """Returns the default hashing algorithm for the Asset Manifest, represented as a string"""
        return DEFAULT_HASH_ALG

    def encode(self) -> str:
        """
        Return a canonicalized JSON string of the manifest
        """
        self.paths.sort(key=canonical_path_comparator)
        return manifest_to_canonical_json_string(manifest=self)


class ManifestModel(BaseManifestModel):
    """
    The asset manifest model for v2023-03-03
    """

    manifest_version: ManifestVersion = ManifestVersion.v2023_03_03
    AssetManifest: Type[AssetManifest] = AssetManifest
    Path: Type[ManifestPath] = ManifestPath



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
All the `deadline farm` commands.
"""

import click
from botocore.exceptions import ClientError  # type: ignore[import]

from ... import api
from ...config import config_file
from ...exceptions import DeadlineOperationError
from .._common import _apply_cli_options_to_config, _cli_object_repr, _handle_error
from .._main import main


@main.group(name="farm")
@_handle_error
def cli_farm():
    """
    Commands to work with farms.
    """


@cli_farm.command(name="list")
@click.option("--profile", help="The AWS profile to use.")
@_handle_error
def farm_list(**args):
    """
    Lists the available farms.
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(**args)

    try:
        response = api.list_farms(config=config)
    except ClientError as exc:
        raise DeadlineOperationError(f"Failed to get Farms from Deadline:\n{exc}") from exc

    # Select which fields to print and in which order
    structured_farm_list = [
        {field: farm[field] for field in ["farmId", "displayName"]} for farm in response["farms"]
    ]

    click.echo(_cli_object_repr(structured_farm_list))


@cli_farm.command(name="get")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@_handle_error
def farm_get(**args):
    """
    Get the details of a farm.

    If farm ID is not provided, returns the configured default farm.
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(required_options={"farm_id"}, **args)

    farm_id = config_file.get_setting("defaults.farm_id", config=config)

    deadline = api.get_boto3_client("deadline", config=config)
    response = deadline.get_farm(farmId=farm_id)
    response.pop("ResponseMetadata", None)

    click.echo(_cli_object_repr(response))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Functionality common to all the CLI groups.
"""

__all__ = [
    "_PROMPT_WHEN_COMPLETE",
    "_prompt_at_completion",
    "_handle_error",
    "_apply_cli_options_to_config",
    "_cli_object_repr",
    "_ProgressBarCallbackManager",
]

import sys
from configparser import ConfigParser
from typing import Any, Callable, Optional, Set
import logging
import traceback

import click
from contextlib import ExitStack
from deadline.job_attachments.progress_tracker import ProgressReportMetadata

from ..config import config_file
from ..exceptions import DeadlineOperationError
from ..job_bundle import deadline_yaml_dump
from ._groups._sigint_handler import SigIntHandler

logger = logging.getLogger("deadline.client.cli")

_PROMPT_WHEN_COMPLETE = "PROMPT_WHEN_COMPLETE"

# Set up the signal handler for handling Ctrl + C interruptions.
sigint_handler = SigIntHandler()


def _prompt_at_completion(ctx: click.Context):
    """
    If the click context has PROMPT_WHEN_COMPLETE set to True,
    prints out a prompt and waits for keyboard input.
    """
    if ctx.obj[_PROMPT_WHEN_COMPLETE]:
        click.prompt(
            "Press Enter To Exit", prompt_suffix="", show_default=False, hide_input=True, default=""
        )


def _handle_error(func: Callable) -> Callable:
    """
    Decorator that catches any exceptions raised in the passed in function,
    and handles their default printout.
    """

    @click.pass_context
    def wraps(ctx: click.Context, *args, **kwargs):
        try:
            func(*args, **kwargs)
        except DeadlineOperationError as e:
            # The message from DeadlineOperationError is printed
            # out verbatim.
            click.echo(str(e))
            _prompt_at_completion(ctx)
            sys.exit(1)
        except click.ClickException:
            # Let click exceptions fall through
            raise
        except Exception:
            # Log and print out unfamiliar exceptions with additional
            # messaging.
            click.echo("The AWS Deadline Cloud CLI encountered the following exception:")
            click.echo(traceback.format_exc())
            _prompt_at_completion(ctx)
            sys.exit(1)

    wraps.__doc__ = func.__doc__
    return wraps


def _apply_cli_options_to_config(
    *, config: Optional[ConfigParser] = None, required_options: Set[str] = set(), **args
) -> Optional[ConfigParser]:
    """
    Modifies an AWS Deadline Cloud config object to apply standard option names to it, such as
    the AWS profile, AWS Deadline Cloud Farm, or AWS Deadline Cloud Queue to use.

    Args:
        config (ConfigParser, optional): an AWS Deadline Cloud config, read by config_file.read_config().
                If not provided, loads the config from disk.
    """
    # Only work with a custom config if there are standard options provided
    if any(value is not None for value in args.values()):
        if config is None:
            config = config_file.read_config()

        aws_profile_name = args.pop("profile", None)
        if aws_profile_name:
            config_file.set_setting("defaults.aws_profile_name", aws_profile_name, config=config)

        farm_id = args.pop("farm_id", None)
        if farm_id:
            config_file.set_setting("defaults.farm_id", farm_id, config=config)

        queue_id = args.pop("queue_id", None)
        if queue_id:
            config_file.set_setting("defaults.queue_id", queue_id, config=config)

        storage_profile_id = args.pop("storage_profile_id", None)
        if storage_profile_id:
            config_file.set_setting(
                "settings.storage_profile_id", storage_profile_id, config=config
            )

        job_id = args.pop("job_id", None)
        if job_id:
            config_file.set_setting("defaults.job_id", job_id, config=config)

        auto_accept = args.pop("yes", None)
        if auto_accept:
            config_file.set_setting("settings.auto_accept", "true", config=config)

        conflict_resolution = args.pop("conflict_resolution", None)
        if conflict_resolution:
            config_file.set_setting(
                "settings.conflict_resolution", conflict_resolution, config=config
            )
    else:
        # Remove the standard option names from the args list
        for name in ["profile", "farm_id", "queue_id", "job_id", "storage_profile_id"]:
            args.pop(name, None)

    # Check that the required options have values
    if "farm_id" in required_options:
        required_options.remove("farm_id")
        if not config_file.get_setting("defaults.farm_id", config=config):
            raise click.UsageError("Missing '--farm-id' or default Farm ID configuration")

    if "queue_id" in required_options:
        required_options.remove("queue_id")
        if not config_file.get_setting("defaults.queue_id", config=config):
            raise click.UsageError("Missing '--queue-id' or default Queue ID configuration")

    if "job_id" in required_options:
        required_options.remove("job_id")
        if not config_file.get_setting("defaults.job_id", config=config):
            raise click.UsageError("Missing '--job-id' or default Job ID configuration")

    if required_options:
        raise RuntimeError(
            f"Unexpected required AWS Deadline Cloud CLI options: {required_options}"
        )

    if args:
        raise RuntimeError(
            f"Option names {tuple(args.keys())} are not standard AWS Deadline Cloud CLI options, they need special handling"
        )

    return config


def _fix_multiline_strings(obj: Any) -> Any:
    """
    Fixes the multi-line strings in `obj` to end with "\n".
    Returns a new object that has been modified.
    """
    if isinstance(obj, str):
        if "\n" in obj and not obj.endswith("\n"):
            return obj + "\n"
        else:
            return obj
    elif isinstance(obj, list):
        return [_fix_multiline_strings(item) for item in obj]
    elif isinstance(obj, tuple):
        return tuple(_fix_multiline_strings(item) for item in obj)
    elif isinstance(obj, dict):
        return {key: _fix_multiline_strings(value) for key, value in obj.items()}
    elif isinstance(obj, set):
        return {_fix_multiline_strings(item) for item in obj}
    else:
        return obj


def _cli_object_repr(obj: Any) -> str:
    """
    Transforms an API response object into a string, for printing as
    CLI output. This formats the output as YAML, using the "|"-style
    for multi-line strings.
    """
    # If a multi-line string does not end with an "\n", the formatting
    # will not use the "|"-style yaml. We fix that up be modifying such
    # strings to end with "\n".
    obj = _fix_multiline_strings(obj)
    return deadline_yaml_dump(obj)


class _ProgressBarCallbackManager:
    """
    Manages creation, update, and deletion of a progress bar. On first call of the callback, the progress bar is created. The progress bar is closed
    on the final call (100% completion)
    """

    BAR_NOT_CREATED = 0
    BAR_CREATED = 1
    BAR_CLOSED = 2

    def __init__(self, length: int, label: str):
        self._length = length
        self._label = label
        self._bar_status = self.BAR_NOT_CREATED
        self._exit_stack = ExitStack()

    def callback(self, upload_metadata: ProgressReportMetadata) -> bool:
        if self._bar_status == self.BAR_CLOSED:
            # from multithreaded execution this can be called after completion somtimes.
            return sigint_handler.continue_operation
        elif self._bar_status == self.BAR_NOT_CREATED:
            # Note: click doesn't export the return type of progressbar(), so we suppress mypy warnings for
            # not annotating the type of hashing_progress.
            self._upload_progress = click.progressbar(length=self._length, label=self._label)  # type: ignore[var-annotated]
            self._exit_stack.enter_context(self._upload_progress)
            self._bar_status = self.BAR_CREATED

        total_progress = int(upload_metadata.progress)
        new_progress = total_progress - self._upload_progress.pos
        if new_progress > 0:
            self._upload_progress.update(new_progress)

        if total_progress == self._length or not sigint_handler.continue_operation:
            self._bar_status = self.BAR_CLOSED
            self._exit_stack.close()

        return sigint_handler.continue_operation



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

# Only needed for access to command line arguments
import os
import signal
import sys
from logging import getLogger
from pathlib import Path

from qtpy.QtCore import Qt
from qtpy.QtGui import QColor, QIcon, QPalette
from qtpy.QtWidgets import QApplication, QFileDialog, QMainWindow, QStyleFactory

from .. import api
from .cli_job_submitter import show_cli_job_submitter
from .dialogs import DeadlineConfigDialog, DeadlineLoginDialog
from .job_bundle_submitter import show_job_bundle_submitter

logger = getLogger(__name__)


class DevMainWindow(QMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.mainColor = QColor(64, 64, 64)
        self.selectionColor = QColor(37, 200, 25)
        self.setMinimumSize(400, 400)
        self.setup_ui()

        signal.signal(signal.SIGINT, self.signal_handler)
        if os.name != "nt":
            signal.signal(signal.SIGQUIT, self.signal_handler)  # type: ignore[attr-defined]
            signal.signal(signal.SIGTERM, self.signal_handler)  # type: ignore[attr-defined]

        # Remove the central widget. This leaves us with just dockable widgets, which provides
        # the most flexibility, since we don't really have a "main" widget.
        self.setCentralWidget(None)  # type: ignore[arg-type]

        self.setDockOptions(
            QMainWindow.AllowNestedDocks | QMainWindow.AllowTabbedDocks | QMainWindow.AnimatedDocks
        )

    def setup_ui(self):
        submit = self.menuBar().addMenu("&Submit job")
        submit.addAction("Submit job bundle...", self.submit_job_bundle)
        submit.addAction("Submit CLI job...", self.submit_cli_job)
        account = self.menuBar().addMenu("&Account")
        account.addAction("AWS Deadline Cloud workstation configuration...", self.configure)
        account.addAction("Log in to AWS Deadline Cloud...", self.login)
        account.addAction("Log out of AWS Deadline Cloud...", self.logout)

        # Set up status bar
        # self.statusBar().setStyleSheet("QStatusBar::item{ border: none; }")
        self.statusBar().showMessage("Testing!")

    def submit_job_bundle(self):
        input_job_bundle_dir = os.path.normpath(
            os.path.join(__file__, "../resources/cli_job_bundle")
        )
        input_job_bundle_dir = QFileDialog.getExistingDirectory(
            self, "Choose job bundle directory", input_job_bundle_dir
        )
        if input_job_bundle_dir:
            show_job_bundle_submitter(
                input_job_bundle_dir=os.path.normpath(input_job_bundle_dir),
                browse=True,
                parent=self,
                f=Qt.Tool,
            )

    def submit_cli_job(self):
        show_cli_job_submitter(self, f=Qt.Tool)

    def configure(self):
        DeadlineConfigDialog.configure_settings(parent=self)

    def login(self):
        if DeadlineLoginDialog.login(parent=self):
            logger.info("Logged in successfully")
        else:
            logger.info("Failed to log in")

    def logout(self):
        api.logout()

    def signal_handler(self, signal, frame):
        self.close()


def app() -> None:
    app = QApplication(sys.argv)

    # Set the style.
    app.setStyle(QStyleFactory.create("fusion"))

    # Set the application info.
    app.setApplicationName("AWS Deadline Cloud client test GUI")
    app.setOrganizationName("AWS")
    app.setOrganizationDomain("https://aws.amazon.com/")
    icon = QIcon(str(Path(__file__).parent / "resources" / "deadline_logo.svg"))
    app.setWindowIcon(icon)

    # Apply the stylesheet.
    pal = QPalette(QColor(64, 64, 64))
    pal.setColor(QPalette.Highlight, QColor(37, 200, 25))
    pal.setColor(QPalette.HighlightedText, QColor(0, 0, 0))
    pal.setColor(QPalette.Link, QColor(96, 185, 250))
    app.setPalette(pal)

    window = app.palette().color(QPalette.Window)
    selection = app.palette().color(QPalette.Highlight)
    app.setStyleSheet(
        """
        *{ selection-background-color: rgb("""
        + str(selection.red())
        + """, """
        + str(selection.green())
        + """, """
        + str(selection.blue())
        + """); }
        QMenu {background-color: rgb("""
        + str(window.red())
        + """, """
        + str(window.green())
        + """, """
        + str(window.blue())
        + """); menu-scrollable: 1;}
        QToolTip{ color: black;}
        QDockWidget {titlebar-close-icon: url(:/ThinkboxUI/Bitmaps/Close2_Dark.png);
                     titlebar-normal-icon: url(:/ThinkboxUI/Bitmaps/Undock2_Dark.png);}
        QDockWidget::close-button, QDockWidget::float-button {min-width: 18px; min-height: 18px; icon-size: 12px;}
        QDockWidget::float-button {position: relative; right: 20px; top: 2px;}
        QDockWidget::close-button {position: relative; right: 1px; top: 2px;}
        QDockWidget::title { text-align: center;}
        QDockWidget::title {background: rgb("""
        + str(max(0, (window.red() - 10)))
        + """, """
        + str(max(0, (window.green() - 10)))
        + """, """
        + str(max(0, (window.blue() - 10)))
        + """);}"""
    )

    main_window = DevMainWindow()
    main_window.show()

    main_window.submit_job_bundle()

    app.exec_()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from typing import Dict, Any
from dataclasses import dataclass
from datetime import timedelta

from ...exceptions import NonValidInputError


@dataclass
class TimeoutEntry:
    """
    Represents a single timeout configuration entry.

    Attributes:
        tooltip (str): Description or explanation of the timeout entry.
        is_activated (bool): Flag indicating if the timeout is activated. Defaults to True.
        seconds (int): Timeout duration in seconds. Defaults to 24 hours (86400 seconds).

    Raises:
        ValueError: If timeout seconds is <= 0 when activated or tooltip is empty.
    """

    tooltip: str
    is_activated: bool = True
    seconds: int = int(timedelta(days=1).total_seconds())

    def __post_init__(self) -> None:
        """Validates the timeout entry after initialization."""
        if self.is_activated and self.seconds <= 0:
            raise ValueError(f"Timeout value cannot be negative or zero: {self.seconds}")
        if not self.tooltip:
            raise ValueError("Timeout tooltip cannot be empty.")

    def to_sticky_settings_dict(self) -> Dict[str, Any]:
        """
        Returns the timeout entry as a dictionary suitable for sticky settings.

        The returned dictionary contains only the essential attributes needed for saving:
        - is_activated: The activation status
        - seconds: The timeout duration

        The tooltip is intentionally excluded as it's not necessary for sticky settings.

        Returns:
            Dict[str, Any]: Dictionary containing activation status and timeout duration.
        """
        return {"is_activated": self.is_activated, "seconds": self.seconds}


@dataclass
class TimeoutTableEntries:
    """
    Manages a collection of named timeout entries.

    This class provides functionality to handle multiple TimeoutEntry instances,
    including serialization to/from sticky settings and validation operations.
    It maintains a mapping of labels to timeout configurations and ensures
    all timeout values are valid when activated.

    Attributes:
        entries (Dict[str, TimeoutEntry]): Dictionary mapping labels to TimeoutEntry instances.
    """

    entries: Dict[str, TimeoutEntry]

    def to_sticky_settings_dict(self) -> Dict[str, Dict[str, Any]]:
        """
        Returns all timeout entries as a dictionary format suitable for sticky settings.

        Returns:
            Dict[str, Dict[str, Any]]: A dictionary mapping labels to entry configurations.
                Example: {
                    "Env1 enter": {"is_activated": True, "seconds": 300},
                    "Env1 exit": {"is_activated": True, "seconds": 60},
                    ...
                }
        """
        return {label: entry.to_sticky_settings_dict() for label, entry in self.entries.items()}

    def update_from_sticky_settings(self, data: Dict[str, Dict[str, Any]]) -> None:
        """
        Updates existing timeout entries with values from sticky settings.

        This method performs a selective update where:
        - Only existing entries are updated (unknown labels in input data are ignored)
        - Missing attributes preserve their current values. So adding a new timeout is
            still a backwards compatible change.

        Args:
            data (Dict[str, Dict[str, Any]]): Dictionary containing timeout configurations.
        """
        for label, saved_data in data.items():
            if label in self.entries:
                entry = self.entries[label]
                entry.is_activated = saved_data.get("is_activated", entry.is_activated)
                entry.seconds = saved_data.get("seconds", entry.seconds)

    def validate_entries(self) -> None:
        """
        Validates all timeout entries in the table.

        Performs validation checks for all entries:
        - Identifies any activated entries with zero duration
        - Collects labels of non valid entries
        - Provides detailed error message for configuration correction

        Raises:
            NonValidInputError: If any activated timeout has a zero duration. The error message
                includes the specific labels of non valid entries and instructions for correction.
        """
        zero_timeouts = [
            label
            for label, entry in self.entries.items()
            if entry.is_activated and entry.seconds == 0
        ]

        if zero_timeouts:
            msg = "The following timeout value(s) must be greater than 0: \n"
            msg += ", ".join(zero_timeouts)
            msg += "\n\nPlease configure these value(s) in the 'Job specific settings' tab."
            raise NonValidInputError(msg)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
This module exists to work around circular imports between the main CLI and each CLI group
"""

from ._main import main  # noqa: F401

# New groups must be imported here to get added to the main CLI as subcommands
from ._groups import (  # noqa: F401
    bundle_group,
    config_group,
    auth_group,
    farm_group,
    fleet_group,
    handle_web_url_command,
    job_group,
    queue_group,
    worker_group,
    attachment_group,
    manifest_group,
)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Contains dataclasses for holding UI parameter values, used by the widgets.
"""

from __future__ import annotations

import os
from dataclasses import dataclass, field
from typing import Literal, Dict, List, Union

from ...job_bundle.parameters import JobParameter

MAX_INT_VALUE = (2**31) - 1


@dataclass
class JobBundleSettings:  # pylint: disable=too-many-instance-attributes
    """
    Settings for the Job Bundle submitter dialog.
    """

    # Used in UI elements and when creating the job bundle directory
    submitter_name: str = field(default="JobBundle")

    # Shared settings
    name: str = field(default="Job bundle")
    description: str = field(default="")

    # Job Bundle settings
    input_job_bundle_dir: str = field(default="")
    parameters: list[JobParameter] = field(default_factory=list)

    # Whether to allow ability to "Load a different job bundle"
    browse_enabled: bool = field(default=False)


@dataclass
class CliJobSettings:  # pylint: disable=too-many-instance-attributes
    """
    Settings for a CLI Job.
    """

    # Used in UI elements and when creating the job bundle directory
    submitter_name: str = field(default="CLI")

    # Shared settings
    name: str = field(default="CLI job")
    description: str = field(default="")

    # CLI job settings
    bash_script_contents: str = field(
        default="""#!/usr/bin/env bash
echo "Data Dir is {{Param.DataDir}}"
cd "{{Param.DataDir}}"

echo "The file contents attached to this job:"
ls

echo "Running index {{Task.Param.Index}}"
sleep 35

# Generate an output file for this task
echo "Content for generated file {{Task.Param.Index}}" > task_output_file_{{Task.Param.Index}}.txt
"""
    )
    use_array_parameter: bool = field(default=True)
    array_parameter_name: str = field(default="Index")
    array_parameter_values: str = field(default="1-5")
    data_dir: str = field(default=os.path.join("~", "CLIJobData"))
    file_format: str = field(default="YAML")


@dataclass
class OsRequirements:
    """
    Settings for a OSRequirementsWidget.
    """

    LINUX = "linux"
    MACOS = "macos"
    WINDOWS = "windows"
    OPERATING_SYSTEMS = [LINUX, MACOS, WINDOWS]
    X86_64 = "x86_64"
    ARM_64 = "arm64"
    CPU_ARCHITECTURES = [X86_64, ARM_64]
    operating_systems: List[Literal[LINUX, MACOS, WINDOWS]] = field(default_factory=list)  # type: ignore
    cpu_archs: List[Literal[X86_64, ARM_64]] = field(default_factory=list)  # type: ignore

    def __post_init__(self):
        for operating_system in self.operating_systems:
            if operating_system not in self.OPERATING_SYSTEMS:
                raise ValueError(
                    f"Operating system {operating_system} is not in supported list: {self.OPERATING_SYSTEMS}."
                )
        for cpu in self.cpu_archs:
            if cpu not in self.CPU_ARCHITECTURES:
                raise ValueError(
                    f"CPU architecture {cpu} is in supported list: {self.CPU_ARCHITECTURES}."
                )

    def serialize(self) -> list:
        requirements: List[dict] = []
        if self.operating_systems:
            for operating_system in self.operating_systems:
                if operating_system not in self.OPERATING_SYSTEMS:
                    raise ValueError(
                        f"Operating system {operating_system} is not in supported list: {self.OPERATING_SYSTEMS}."
                    )
            requirements.append(
                {
                    "name": "attr.worker.os.family",
                    "anyOf": self.operating_systems,
                }
            )
        if self.cpu_archs:
            for cpu in self.cpu_archs:
                if cpu not in self.CPU_ARCHITECTURES:
                    raise ValueError(
                        f"CPU architecture {cpu} is not in supported list: {self.CPU_ARCHITECTURES}."
                    )
            requirements.append(
                {
                    "name": "attr.worker.cpu.arch",
                    "anyOf": self.cpu_archs,
                }
            )
        return requirements


@dataclass
class HardwareRequirements:
    """
    Settings for a HardwareRequirementsWidget.
    """

    DEFAULT_VALUE = -1
    cpu_min: int = field(default=DEFAULT_VALUE)
    cpu_max: int = field(default=DEFAULT_VALUE)
    memory_min: int = field(default=DEFAULT_VALUE)
    memory_max: int = field(default=DEFAULT_VALUE)
    acceleration_min: int = field(default=DEFAULT_VALUE)
    acceleration_max: int = field(default=DEFAULT_VALUE)
    acceleration_memory_min: int = field(default=DEFAULT_VALUE)
    acceleration_memory_max: int = field(default=DEFAULT_VALUE)
    scratch_space_min: int = field(default=DEFAULT_VALUE)
    scratch_space_max: int = field(default=DEFAULT_VALUE)

    def __post_init__(self):
        self._validate(self.cpu_min, self.cpu_max, "CPU")
        self._validate(self.memory_min, self.memory_max, "Memory")
        self._validate(self.acceleration_min, self.acceleration_max, "Acceleration")
        self._validate(
            self.acceleration_memory_min, self.acceleration_memory_max, "Acceleration Memory"
        )
        self._validate(self.scratch_space_min, self.scratch_space_max, "Scratch Space")

    def _validate(self, minimum: int, maximum: int, name: str):
        if minimum != self.DEFAULT_VALUE and maximum != self.DEFAULT_VALUE and minimum > maximum:
            raise ValueError(
                f"{name} Minimum cannot be higher than {name} Maximum. {minimum} > {maximum}"
            )

    def _serialize(self, minimum, maximum, name):
        requirements: List[dict] = []
        if minimum != self.DEFAULT_VALUE or maximum != self.DEFAULT_VALUE:
            _requirement = {
                "name": name,
            }
            if minimum != self.DEFAULT_VALUE:
                _requirement["min"] = max(minimum, self.DEFAULT_VALUE)
            if maximum != self.DEFAULT_VALUE:
                _requirement["max"] = min(maximum, MAX_INT_VALUE)
            requirements.append(_requirement)
        return requirements

    def serialize(self) -> list:
        requirements: List[dict] = []
        requirements.extend(self._serialize(self.cpu_min, self.cpu_max, "amount.worker.vcpu"))
        requirements.extend(
            self._serialize(self.memory_min, self.memory_max, "amount.worker.memory")
        )
        requirements.extend(
            self._serialize(self.acceleration_min, self.acceleration_max, "amount.worker.gpu")
        )
        requirements.extend(
            self._serialize(
                self.acceleration_memory_min,
                self.acceleration_memory_max,
                "amount.worker.gpu.memory",
            )
        )
        requirements.extend(
            self._serialize(
                self.scratch_space_min, self.scratch_space_max, "amount.worker.disk.scratch"
            )
        )
        return requirements


@dataclass
class CustomAmountRequirement:
    """
    Settings for a CustomAmountWidget
    """

    DEFAULT_VALUE = -(2**31) + 1
    name: str = field(default="")
    min: int = field(default=DEFAULT_VALUE)
    max: int = field(default=DEFAULT_VALUE)

    def __post_init__(self):
        if not self.name:
            raise ValueError(f"Custom Amount {self} has no name")
        elif self.min and self.max and self.min > self.max:
            raise ValueError(f"Custom Amount {self} has min higher than max")

    def serialize(self) -> dict:
        _requirement = {"name": "amount.worker." + self.name}
        if self.min != self.DEFAULT_VALUE:
            _requirement["min"] = max(self.min, self.DEFAULT_VALUE)
        if self.max != self.DEFAULT_VALUE:
            _requirement["max"] = min(self.max, MAX_INT_VALUE)
        return _requirement


@dataclass
class CustomAttributeRequirement:
    """
    Settings for a CustomAttributeWidget
    """

    ANY_OF = "anyOf"
    ALL_OF = "allOf"
    OPTIONS = [ANY_OF, ALL_OF]
    name: str = field(default="")
    option: Literal[ANY_OF, ALL_OF] = field(default=ALL_OF)  # type: ignore
    values: List[str] = field(default_factory=list)

    def __post_init__(self):
        if not self.name:
            raise ValueError(f"Custom Amount {self} has no name")
        elif not self.option or self.option not in CustomAttributeRequirement.OPTIONS:
            raise ValueError(
                f"Custom Amount {self} option is not in {CustomAttributeRequirement.OPTIONS}"
            )

    def serialize(self) -> dict:
        if not self.name:
            raise ValueError(f"Custom Attribute {self} has no name")
        _requirement = {
            "name": "attr.worker." + self.name,
        }
        if not self.option:
            raise ValueError(f"Custom Attribute {self} has no option")
        elif self.option not in ["anyOf", "allOf"]:
            raise ValueError(
                f'Custom Attribute {self} has option {self.option} which is not in list: ["anyOf", "allOf"]'
            )
        if not self.values:
            raise ValueError(f"Custom Attribute {self} has no values")
        _requirement[self.option] = self.values
        return _requirement


@dataclass
class CustomRequirements:
    """
    Settings for a CustomRequirementsWidget.
    """

    amounts: List[CustomAmountRequirement] = field(default_factory=list)
    attributes: List[CustomAttributeRequirement] = field(default_factory=list)

    def __post_init__(self):
        amounts = []
        attributes = []
        for amount in self.amounts:
            amounts.append(self._validate_amount(amount))
        for attribute in self.attributes:
            attributes.append(self._validate_attribute(attribute))
        self.amounts = amounts
        self.attributes = attributes

    @staticmethod
    def _validate_amount(amount: Union[dict, CustomAmountRequirement]) -> CustomAmountRequirement:
        if isinstance(amount, dict):
            amount = CustomAmountRequirement(**amount)
        return amount

    @staticmethod
    def _validate_attribute(
        attribute: Union[dict, CustomAttributeRequirement],
    ) -> CustomAttributeRequirement:
        if isinstance(attribute, dict):
            attribute = CustomAttributeRequirement(**attribute)
        return attribute

    def __iter__(self):
        if self.amounts:
            yield from self.amounts
        if self.attributes:
            yield from self.attributes

    def _serialize_amounts(self):
        requirements = []
        for amount in self.amounts:
            requirements.append(amount.serialize())
        return requirements

    def _serialize_attributes(self):
        requirements = []
        for attribute in self.attributes:
            requirements.append(attribute.serialize())
        return requirements

    def serialize(self) -> Dict[str, List]:
        requirements: Dict[str, List] = {}

        if self.amounts:
            requirements.setdefault("amounts", []).extend(self._serialize_amounts())

        if self.attributes:
            requirements.setdefault("attributes", []).extend(self._serialize_attributes())
        return requirements


@dataclass
class HostRequirements:
    """
    Settings for a HostRequirementsWidget.
    """

    os_requirements: OsRequirements = field(default=None)
    hardware_requirements: HardwareRequirements = field(default=None)
    custom_requirements: CustomRequirements = field(default=None)

    def __post_init__(self):
        if isinstance(self.os_requirements, dict):
            self.os_requirements = OsRequirements(**self.os_requirements)
        if isinstance(self.hardware_requirements, dict):
            self.hardware_requirements = HardwareRequirements(**self.hardware_requirements)
        if isinstance(self.custom_requirements, dict):
            self.custom_requirements = CustomRequirements(**self.custom_requirements)

    def serialize(self) -> dict:
        requirements: Dict[str, Union[List[str], List[int]]] = {}
        if self.os_requirements is not None:
            requirements.setdefault("attributes", []).extend(self.os_requirements.serialize())

        if self.hardware_requirements:
            requirements.setdefault("amounts", []).extend(self.hardware_requirements.serialize())

        custom_requirements = self.custom_requirements.serialize()
        if custom_requirements.get("amounts", []):
            requirements.setdefault("amounts", []).extend(custom_requirements["amounts"])
        if custom_requirements.get("attributes", []):
            requirements.setdefault("attributes", []).extend(custom_requirements["attributes"])

        return requirements



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
from contextlib import contextmanager
from typing import Any

from ..exceptions import DeadlineOperationError


@contextmanager
def block_signals(element):
    """
    Context manager used to turn off signals for a UI element.
    """
    old_value = element.blockSignals(True)
    try:
        yield
    finally:
        element.blockSignals(old_value)


@contextmanager
def gui_error_handler(message_title: str, parent: Any = None):
    """
    A context manager that initializes a Qt GUI context that
    catches errors and shows them in a message box instead of
    punting them to a CLI interface.

    For example:

    with gui_context():
        from deadline.client.ui.cli_job_submitter import show_cli_job_submitter

        show_cli_job_submitter()

    """
    try:
        from qtpy.QtWidgets import QMessageBox

        yield
    except DeadlineOperationError as e:
        QMessageBox.warning(parent, message_title, str(e))  # type: ignore[call-arg]
    except Exception:
        import traceback

        QMessageBox.warning(parent, message_title, f"Exception caught:\n{traceback.format_exc()}")  # type: ignore[call-arg]


@contextmanager
def gui_context_for_cli(automatically_install_dependencies: bool):
    """
    A context manager that initializes a Qt GUI context for
    the CLI handler to use.

    For example:

    with gui_context_for_cli(automatically_install_dependencies) as app:
        from deadline.client.ui.cli_job_submitter import show_cli_job_submitter

        show_cli_job_submitter()

        app.exec()

    If automatically_install_dependencies is true, dependencies will be installed without prompting the user. Useful
    if the command is triggered not from a command line.
    """
    import importlib
    import os
    from os.path import basename, dirname, join, normpath
    import shlex
    import shutil
    import subprocess
    import sys
    from pathlib import Path

    import click

    has_pyside6 = importlib.util.find_spec("PySide6")
    has_pyside2 = importlib.util.find_spec("PySide2")
    if not (has_pyside6 or has_pyside2):
        if not automatically_install_dependencies:
            message = "Optional GUI components for deadline are unavailable. Would you like to install PySide?"
            will_install_gui = click.confirm(message, default=False)
            if not will_install_gui:
                click.echo("Unable to continue without GUI, exiting")
                sys.exit(1)

        # this should match what's in the pyproject.toml
        pyside6_pypi = "PySide6-essentials >= 6.6,< 6.9"
        if "deadline" in basename(sys.executable).lower():
            # running with a deadline executable, not standard python.
            # So exit the deadline folder into the main deps dir
            deps_folder = normpath(
                join(
                    dirname(__file__),
                    "..",
                    "..",
                    "..",
                )
            )
            runtime_version = f"{sys.version_info.major}.{sys.version_info.minor}"
            pip_command = [
                "-m",
                "pip",
                "install",
                pyside6_pypi,
                "--python-version",
                runtime_version,
                "--only-binary=:all:",
                "-t",
                deps_folder,
            ]
            python_executable = shutil.which("python3") or shutil.which("python")
            if python_executable:
                command = " ".join(shlex.quote(v) for v in [python_executable] + pip_command)
                subprocess.run([python_executable] + pip_command)
            else:
                click.echo(
                    "Unable to install GUI dependencies, if you have python available you can install it by running:"
                )
                click.echo()
                click.echo(f"\t{' '.join(shlex.quote(v) for v in ['python'] + pip_command)}")
                click.echo()
                sys.exit(1)
        else:
            # standard python sys.executable
            # TODO: swap to deadline[gui]==version once published and at the same
            # time consider local editables `pip install .[gui]`
            subprocess.run([sys.executable, "-m", "pip", "install", pyside6_pypi])

    # set QT_API to inform qtpy which dependencies to look for.
    # default to pyside6 and fallback to pyside2.
    # Does not work with PyQt5 which is qtpy default
    os.environ["QT_API"] = "pyside6"
    if has_pyside2:
        os.environ["QT_API"] = "pyside2"
    try:
        from qtpy.QtGui import QIcon
        from qtpy.QtWidgets import QApplication, QMessageBox
    except ImportError as e:
        click.echo(f"Failed to import qtpy/PySide/Qt, which is required to show the GUI:\n{e}")
        sys.exit(1)

    try:
        app = QApplication(sys.argv)
        app.setApplicationName("AWS Deadline Cloud")
        icon = QIcon(str(Path(__file__).parent.parent / "ui" / "resources" / "deadline_logo.svg"))
        app.setWindowIcon(icon)

        yield app
    except DeadlineOperationError as e:
        import os
        import shlex

        command = f"{os.path.basename(sys.argv[0])} " + " ".join(
            shlex.quote(v) for v in sys.argv[1:]
        )
        QMessageBox.warning(None, f'Error running "{command}"', str(e))  # type: ignore[call-overload, call-arg, arg-type]
    except Exception:
        import os
        import shlex
        import traceback

        command = f"{os.path.basename(sys.argv[0])} " + " ".join(
            shlex.quote(v) for v in sys.argv[1:]
        )
        QMessageBox.warning(  # type: ignore[call-overload, call-arg]
            None,  # type: ignore[arg-type]
            f'Error running "{command}"',
            f"Exception caught:\n{traceback.format_exc()}",
        )


class CancelationFlag:
    """
    Helper object for background thread cancelation.
    The `destroyed` event cannot be connected to a member
    function of the class. With this object, you can bind it
    to the cancelation flag's set_canceled method instead.

    Example usage:

    class MyWidget(QWidget):
        thread_event = Signal(str)

        def __init__(self):
            self.canceled = CancelationFlag()
            self.destroyed.connect(self.canceled.set_canceled)

        def _my_thread_function(self):
            ...<processing>...

            if not self.canceled:
                self.thread_event.emit(result)
    """

    def __init__(self):
        self.canceled = False

    def set_canceled(self):
        self.canceled = True

    def __bool__(self):
        return self.canceled



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
import sys
import psutil
from contextlib import contextmanager
from ..exceptions import DeadlineOperationError
import logging

__all__ = ["PidFileLock", "PidLockAlreadyHeld"]

logger = logging.getLogger(__name__)


class PidLockAlreadyHeld(DeadlineOperationError):
    """Error for when the pid lock is already by a process"""


def _pid_lock_temp_file_path(pid_file_path: str) -> str:
    """Construct the temporary file path used to populate the pid lock file before moving it to the lock file path."""
    return pid_file_path + f"{os.getpid()}~tmp"


def _claim_pid_lock_with_rename(tmp_file_name: str, pid_file_path: str) -> bool:
    # Atomic rename the temporary location to the pid lock file. This operation needs to be
    # atomic: it either succeeds or another process is holding the lock.
    #
    # If the lock is not claimed, the temporary file will remain.
    try:
        if sys.platform.startswith("win32") or sys.platform.startswith("cygwin"):
            # On Windows systems, rename will raise if the destination file exists
            os.rename(tmp_file_name, pid_file_path)
        else:
            # On POSIX systems, link will raise if the destination file exists
            os.link(tmp_file_name, pid_file_path)
            os.remove(tmp_file_name)

        # Successfully claimed the lock
        return True
    except FileExistsError:
        # Another process was holding the lock, as the file exists.
        return False


def _try_acquire_pid_lock(
    pid_file_path: str,
    operation_name: str = "the operation",
):
    """
    Checks if the specified pid lock file exists and executes as per the following:
    If the pid lock file does not exist:
        It creates a new pid file and acquires lock for this pid.
        It handles concurrent processes trying to obtain lock such that only one process gets the lock and others fail

    If the pid lock file exists:
        If the process with its pid is still running, it raises an exception that a download is in progress already
        If the process with its pid is not running it deletes the pid lock file after taking a primitive file lock on it
        to handle concurrent processes making the same check. This will not work if primitive file locks are disabled.

    :param pid_file_full_path: full path of the pid lock file
    :return: boolean, True if pid lock was obtained successfully, throws an exception otherwise
    """
    current_process_id: int = os.getpid()

    # Generate a tmp file for writing the pid file as a whole and prevent corrupt data
    tmp_file_path = _pid_lock_temp_file_path(pid_file_path)

    try:
        # Write the pid lock file to the temporary location
        with open(tmp_file_path, "w+") as f:
            f.write(str(current_process_id))

        # Claim the lock if possible. We do this first so that the normal path does the fewest operations.
        if _claim_pid_lock_with_rename(tmp_file_path, pid_file_path):
            return

        # If we could not claim the lock, inspect it to see whether it's stale and should be deleted,
        # for example if a previous run of the command was terminated from task manager and could not clean up.
        try:
            # Read the lock file. If it does not exist, this raises FileNotFoundError.
            with open(pid_file_path, "r") as f:
                try:
                    lock_holder_pid = int(f.read())
                except ValueError:
                    lock_holder_pid = -1
            # If the lock holder PID is not running, that means it exited without a proper shutdown,
            # and we should clear the lock. Under normal operation this will not occur. There is
            # a race if another process is running and discovers this at the same time. To minimize
            # the time window of the race, reading the lock, checking the pid, and removing the file are
            # done in sequence.
            if lock_holder_pid == -1:
                os.remove(pid_file_path)
                logger.warning("Pid lock file contains incorrect data. Deleted pid lock file.")
            elif not psutil.pid_exists(lock_holder_pid):
                os.remove(pid_file_path)
                logger.warning(
                    f"Process with pid {lock_holder_pid} is not running. Deleted pid lock file."
                )
            else:
                raise PidLockAlreadyHeld(
                    f"Unable to perform {operation_name} as process with pid {lock_holder_pid} already holds the lock {pid_file_path}"
                )
        except FileNotFoundError:
            # In this case, the pid lock is free to acquire
            pass

        # After possibly cleaning up a stale lock, try claiming it again
        if _claim_pid_lock_with_rename(tmp_file_path, pid_file_path):
            return
        else:
            raise PidLockAlreadyHeld(
                f"Unable to perform {operation_name} as process with pid {lock_holder_pid} already holds the lock {pid_file_path}"
            )

    finally:
        # Clean up the pid lock temporary file if necessary
        if os.path.exists(tmp_file_path):
            try:
                os.remove(tmp_file_path)
            except OSError as e:
                logger.warning(f"Failed to clean up pid lock temporary file: {e}")


def _release_pid_lock(pid_file_path: str):
    """
    Releases the pid lock by deleting the pid file.

    :param pid_file_full_path: full path of the pid lock file
    :return: boolean, True if pid lock released successfully
    """
    # Get the current process's id to obtain lock
    current_process_id: int = os.getpid()

    # If an issue occurred with the temporary file during acquisition, delete it here.
    tmp_file_name = _pid_lock_temp_file_path(pid_file_path)
    if os.path.exists(tmp_file_name):
        try:
            os.remove(tmp_file_name)
            logger.warning(f"Cleaned up stale pid lock temporary file: {tmp_file_name}")
        except OSError as e:
            logger.warning(f"Failed to clean up pid lock temporary file: {e}")

    # Check if pid lock file does not exist. Returns if file doesn't exist as there is no lock to be released.
    if not os.path.exists(pid_file_path):
        logger.warning(f"Expected pid lock file does not exist at {pid_file_path}")
        return

    # Try to open pid file at download progress location in read mode
    with open(pid_file_path, "r") as f:
        # Read pid file and obtain the process id from file contents
        lock_holder_pid = f.read()

    if lock_holder_pid == str(current_process_id):
        # Process pid from file is same as current process pid - release pid lock
        os.remove(pid_file_path)
    else:
        # Process pid from file is different from current process pid.
        logger.warning(
            f"Another process with pid {lock_holder_pid} claimed the pid lock {pid_file_path} while {current_process_id} was holding it. Skipping pid file deletion."
        )


@contextmanager
def PidFileLock(
    lock_file_path: str,
    operation_name: str = "the operation",
):
    """
    A context manager for holding a pid (process id) lock file during the scope of a 'with' statement.
    A pid lock file lets you prevent concurrent execution of the same CLI command. For example,
    a command to repeatedly download the new output available from a Deadline Cloud queue could
    use this to ensure only one running command is calculating and downloading what to output at a time.

    Example:
        with PidLockFile("/path/to/lock/file", operation_name="incremental output download"):
            # Code to load the checkpoint, do the download, save the new checkpoint...

    Args:
        lock_file_path (str): The file system path of the PID lock file.
        operation_name (Optional[str]): The name of the operation being performed in the lock, used for error messages.
    """
    _try_acquire_pid_lock(lock_file_path, operation_name)
    try:
        yield None
    finally:
        _release_pid_lock(lock_file_path)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

__all__ = [
    "get_config_file_path",
    "get_cache_directory",
    "read_config",
    "write_config",
    "get_setting_default",
    "get_setting",
    "set_setting",
    "clear_setting",
    "get_best_profile_for_farm",
    "str2bool",
    "DEFAULT_QUEUE_INCREMENTAL_DOWNLOAD_DIR",
]

import getpass
import os
import platform
from configparser import ConfigParser
from pathlib import Path
from typing import Any, Dict, List, Optional
import tempfile

import boto3
from deadline.job_attachments.models import FileConflictResolution

from ..exceptions import DeadlineOperationError

# Default path where AWS Deadline Cloud's configuration lives
CONFIG_FILE_PATH = os.path.join("~", ".deadline", "config")
# Environment variable that, if set, overrides the value of CONFIG_FILE_PATH
CONFIG_FILE_PATH_ENV_VAR = "DEADLINE_CONFIG_FILE_PATH"
# The default AWS Deadline Cloud endpoint URL
# Environment variable that, if set, overrides the value of DEFAULT_DEADLINE_ENDPOINT_URL
DEFAULT_DEADLINE_ENDPOINT_URL = os.getenv(
    "AWS_ENDPOINT_URL_DEADLINE",
    f"https://deadline.{boto3.Session().region_name}.amazonaws.com",
)

# Default directories used by various Deadline CLI commands
DEFAULT_JOB_HISTORY_DIR = os.path.join("~", ".deadline", "job_history", "{aws_profile_name}")
DEFAULT_CACHE_DIR = os.path.join("~", ".deadline", "cache")
DEFAULT_QUEUE_INCREMENTAL_DOWNLOAD_DIR = os.path.join("~", ".deadline", "incremental_download")

_TRUE_VALUES = {"yes", "on", "true", "1"}
_FALSE_VALUES = {"no", "off", "false", "0"}
_BOOL_VALUES = _TRUE_VALUES | _FALSE_VALUES

__config = ConfigParser()
__config_file_path = None
__config_mtime = None

# This value defines the AWS Deadline Cloud settings structure. For each named setting,
# it stores:
# "default" - The default value.
# "depend"  - The setting it depends on, if any. This modifies the section name of the
#             setting to embed the dependency value, e.g. default.farm_id goes in
#             section [profile-{profileName} default]
# "section_format" - How its value gets formatted into config file sections.
SETTINGS: Dict[str, Dict[str, Any]] = {
    "deadline-cloud-monitor.path": {
        "default": "",
        "description": "The filesystem path to Deadline Cloud monitor, set during login process.",
    },
    "defaults.aws_profile_name": {
        "default": "(default)",
        "section_format": "profile-{}",
        "description": "The AWS profile name to use by default. Set to '' to use the default credentials."
        + " Other settings are saved with the profile.",
    },
    "settings.job_history_dir": {
        "default": DEFAULT_JOB_HISTORY_DIR,
        "depend": "defaults.aws_profile_name",
        "description": "The directory in which to place the job submission history for this AWS profile name.",
    },
    "defaults.farm_id": {
        "default": "",
        "depend": "defaults.aws_profile_name",
        "section_format": "{}",
        "description": "The Farm ID to use by default.",
    },
    "settings.storage_profile_id": {
        "default": "",
        "depend": "defaults.farm_id",
        "section_format": "{}",
        "description": "The storage profile that this workstation conforms to. It specifies where shared file systems are mounted, and where named job attachments should go.",
    },
    "defaults.queue_id": {
        "default": "",
        "depend": "defaults.farm_id",
        "section_format": "{}",
        "description": "The Queue ID to use by default.",
    },
    "defaults.job_id": {
        "default": "",
        "depend": "defaults.queue_id",
        "description": "The Job ID to use by default. This gets updated by job submission, so is normally the most recently submitted job.",
    },
    "settings.auto_accept": {
        "default": "false",
        "description": "Automatically accept the default choice for any interactive prompts.",
    },
    "settings.conflict_resolution": {
        "default": FileConflictResolution.NOT_SELECTED.name,
        "description": "How to handle downloads if a file already exists",
    },
    "settings.log_level": {
        "default": "WARNING",
        "description": "The logging level to use in the CLI and GUIs.",
    },
    "telemetry.opt_out": {
        "default": "false",
        "description": "If set to 'true', don't record any telemetry events.",
    },
    "telemetry.identifier": {
        "default": "",
        "description": "A randomly generated identifier used to record telemetry events for this configuration.",
    },
    "defaults.job_attachments_file_system": {
        "default": "COPIED",
        "depend": "defaults.farm_id",
        "description": "The file system mode to use for job attachments when running jobs. COPIED means to download a copy of the attachment data, VIRTUAL means to use a virtual file system for lazy loading.",
    },
    "settings.s3_max_pool_connections": {
        "default": "50",
        "description": (
            "The maximum number of connections to keep in the connection pool used by the S3's upload/download operations. "
            "If this value is not set, the default value of 50 is used. "
            "(Note: It's recommended setting this value above 10 to avoid 'Connection pool is full' warnings during the uploads/downloads.)"
        ),
    },
    "settings.small_file_threshold_multiplier": {
        "default": "20",  # By default, the small file threshold is 160 MB (since the default S3 multipart-upload chunk size is 8 MB.)
        "description": (
            "When uploading job attachments, the file size threshold is set to separate 'large' files from 'small' files so that 'large' files can be processed serially. "
            "This multiplier is used to calculate the size threshold. (Small files are defined as those smaller than or equal to the chunk size multiplied by this factor.)"
        ),
    },
    "settings.known_asset_paths": {
        "default": "",  # OS-specific path separator delimited list
        "description": "A list of paths that should not generate warnings when outside storage profile locations, separated by the OS path list separator (semicolon on Windows, colon on Linux/macOS).",
    },
}


def get_config_file_path() -> Path:
    """
    Get the config file path from the environment variable, falling back
    to our default if it is not set.
    """
    return Path(os.environ.get(CONFIG_FILE_PATH_ENV_VAR) or CONFIG_FILE_PATH).expanduser()


def get_cache_directory() -> str:
    """
    Get the cache directory.
    """
    return os.path.expanduser(DEFAULT_CACHE_DIR)


def _should_read_config(config_file_path: Path) -> bool:
    global __config_file_path
    global __config_mtime

    if (
        __config_mtime is not None
        and config_file_path == __config_file_path
        and config_file_path.is_file()
    ):
        mtime = config_file_path.stat().st_mtime
        if mtime == __config_mtime:
            return False
    return True


def read_config() -> ConfigParser:
    """
    If the config hasn't been read yet, or was modified since it was last
    read, reads the AWS Deadline Cloud configuration.
    """
    global __config
    global __config_file_path
    global __config_mtime

    config_file_path = get_config_file_path()

    if _should_read_config(config_file_path):
        # Read the config file with a fresh config parser, and update the last-modified time stamp
        __config = ConfigParser()
        __config_file_path = config_file_path
        try:
            with open(config_file_path, mode="r", encoding="utf-8") as fh:
                config_contents = fh.read()
            __config.read_string(config_contents, str(config_file_path))
        except FileNotFoundError:
            # If the config file doesn't exist, leave an empty ConfigParser() object, in all
            # other cases let the error pass through.
            pass
        if config_file_path.is_file():
            __config_mtime = config_file_path.stat().st_mtime
        else:
            __config_mtime = None

    return __config


def _reset_directory_permissions_windows(directory: Path) -> None:
    if platform.system() != "Windows":
        return
    import win32security
    import ntsecuritycon

    # We don't want to propagate existing permissions, so create a new DACL
    dacl = win32security.ACL()

    # On Windows, both SYSTEM and the Administrators group normally
    # have Full Access to files in the user's home directory.
    # Use SIDs to represent the Administrators and SYSTEM to
    # support multi-language operating systems
    # Administrator(S-1-5-32-544), SYSTEM(S-1-5-18)
    # https://learn.microsoft.com/en-us/windows/win32/secauthz/well-known-sids
    system_sid = win32security.ConvertStringSidToSid("S-1-5-18")
    admin_sid = win32security.ConvertStringSidToSid("S-1-5-32-544")

    username = getpass.getuser()
    user_sid, _, _ = win32security.LookupAccountName(None, username)

    for sid in [user_sid, admin_sid, system_sid]:
        dacl.AddAccessAllowedAceEx(
            win32security.ACL_REVISION,
            ntsecuritycon.OBJECT_INHERIT_ACE | ntsecuritycon.CONTAINER_INHERIT_ACE,
            ntsecuritycon.GENERIC_ALL,
            sid,
        )

    # Get the security descriptor of the object
    sd = win32security.GetFileSecurity(
        str(directory.resolve()), win32security.DACL_SECURITY_INFORMATION
    )

    # Set the security descriptor's DACL to the newly-created DACL
    # Arguments:
    # 1. bDaclPresent = 1: Indicates that the DACL is present in the security descriptor.
    #    If set to 0, this method ignores the provided DACL and allows access to all principals.
    # 2. dacl: The discretionary access control list (DACL) to be set in the security descriptor.
    # 3. bDaclDefaulted = 0: Indicates the DACL was provided and not defaulted.
    #    If set to 1, indicates the DACL was defaulted, as in the case of permissions inherited from a parent directory.
    sd.SetSecurityDescriptorDacl(1, dacl, 0)
    win32security.SetFileSecurity(
        str(directory.resolve()), win32security.DACL_SECURITY_INFORMATION, sd
    )


def write_config(config: ConfigParser) -> None:
    """
    Writes the provided config to the AWS Deadline Cloud configuration.

    Args:
        config (ConfigParser): The config object to write. Generally this is
            a modified value from what `read_config` returns.
    """
    config_file_path = get_config_file_path()
    if not config_file_path.parent.exists():
        config_file_path.parent.mkdir(parents=True, exist_ok=True)
        if platform.system() == "Windows":
            _reset_directory_permissions_windows(config_file_path.parent)

    # Using the config file path as the prefix ensures that the tmpfile and real file are
    # on the same filesystem. This is a requirement for os.replace to be atomic.
    file_descriptor, tmp_file_name = tempfile.mkstemp(prefix=str(config_file_path), text=True)

    # Note: The file descriptor is closed when exiting the context manager.
    with os.fdopen(file_descriptor, "w", encoding="utf8") as configfile:
        config.write(configfile)

    os.replace(tmp_file_name, config_file_path)


def _get_setting_config(setting_name: str) -> dict:
    """
    Gets the setting configuration for the specified setting name,
    raising an error if it does not exist.

    Args:
        setting_name (str): The full setting name, like `section.setting`.
    """
    setting_config = SETTINGS.get(setting_name)
    if not setting_config:
        raise DeadlineOperationError(
            f"AWS Deadline Cloud configuration has no setting named {setting_name!r}."
        )
    return setting_config


def _get_default_from_setting_config(setting_config, config: Optional[ConfigParser]) -> str:
    """
    Gets the default value from a setting_config entry, performing value substitutions.

    Currently the only substitution supported is `{aws_profile_name}`.
    """
    default_value = setting_config["default"]
    # Only do the substitution if we see a pattern
    if "{" in default_value:
        default_value = default_value.format(
            aws_profile_name=get_setting("defaults.aws_profile_name", config=config)
        )
    return default_value


def get_setting_default(setting_name: str, config: Optional[ConfigParser] = None) -> str:
    """
    Gets the default value for the setting `setting_name`.
    Raises an exception if the setting does not exist.

    Args:
        setting_name (str): The full setting name, like `section.setting`.
        config: The config file read with `read_config()`.
    """
    setting_config = _get_setting_config(setting_name)
    return _get_default_from_setting_config(setting_config, config=config)


def _get_section_prefixes(setting_config: dict, config: ConfigParser) -> List[str]:
    """
    Gets a list of the section name prefixes for the specified setting section + setting config

    Args:
        setting_config: The setting config object from the SETTINGS dictionary.
        config: The config file read with `read_config()`.
    """
    if "depend" in setting_config:
        dep_setting_name = setting_config["depend"]
        dep_setting_config = SETTINGS[dep_setting_name]
        dep_section_format = dep_setting_config["section_format"]

        dep_section, dep_name = dep_setting_name.split(".", 1)
        dep_section_prefixes = _get_section_prefixes(dep_setting_config, config)
        dep_section_value: Optional[str] = config.get(
            " ".join(dep_section_prefixes + [dep_section]), dep_name, fallback=None
        )
        if dep_section_value is None:
            dep_section_value = _get_default_from_setting_config(dep_setting_config, config=config)
        formatted_section_value = dep_section_format.format(dep_section_value)
        return dep_section_prefixes + [formatted_section_value]
    else:
        return []


def get_setting(setting_name: str, config: Optional[ConfigParser] = None) -> str:
    """
    Gets the value of the specified setting, returning the default if
    not configured. Raises an exception if the setting does not exist.

    Args:
        setting_name (str): The full setting name, like `section.setting`.
        config (ConfigParser, optional): The config file read with `read_config()`.
    """
    if "." not in setting_name:
        raise DeadlineOperationError(f"The setting name {setting_name!r} is not valid.")
    section, name = setting_name.split(".", 1)

    if config is None:
        config = read_config()

    setting_config = _get_setting_config(setting_name)

    section_prefixes = _get_section_prefixes(setting_config, config)
    section = " ".join(section_prefixes + [section])

    result: Optional[str] = config.get(section, name, fallback=None)

    if result is None:
        return _get_default_from_setting_config(setting_config, config=config)
    else:
        return result


def set_setting(setting_name: str, value: str, config: Optional[ConfigParser] = None):
    """
    Sets the value of the specified setting, returning the default if
    not configured. Raises an exception if the setting does not exist.

    Args:
        setting_name (str): The full setting name, like `section.setting`.
        value (bool|int|float|str): The value to set.
        config (Optional[ConfigParser]): If provided sets the setting in the parser and does not save to disk.
    """
    if "." not in setting_name:
        raise DeadlineOperationError(f"The setting name {setting_name!r} is not valid.")
    section, name = setting_name.split(".", 1)

    # Get the type of the default to validate it is an AWS Deadline Cloud setting, and retrieve its type
    setting_config = _get_setting_config(setting_name)

    # If no config was provided, then read from disk and signal to write it later
    if not config:
        config = read_config()
        save_config = True
    else:
        save_config = False

    section_prefixes = _get_section_prefixes(setting_config, config)
    section = " ".join(section_prefixes + [section])
    if section not in config:
        config[section] = {}
    config.set(section, name, value)
    if save_config:
        write_config(config)


def clear_setting(setting_name: str, config: Optional[ConfigParser] = None):
    """
    Sets the value of the specified setting back to the default value.

    Args:
        setting_name (str): The full setting name, like `section.setting`.
        config (Optional[ConfigParser]): If provided clears the setting in the parser and does not save to disk.
    """
    if "." not in setting_name:
        raise DeadlineOperationError(f"The setting name {setting_name!r} is not valid.")

    set_setting(setting_name, get_setting_default(setting_name, config=config), config=config)


def get_best_profile_for_farm(farm_id: str, queue_id: Optional[str] = None) -> str:
    """
    Finds the best AWS profile for the specified farm and queue IDs. Chooses
    the first match from:
    1. The default AWS profile if its default farm matches.
    2. AWS profiles whose default farm and queue IDs match.
    3. AWS profiles whose default farm matches.
    4. If there were no matches, returns the default AWS profile.
    """
    # Get the full list of AWS profiles
    session = boto3.Session()
    aws_profile_names = session._session.full_config["profiles"].keys()

    # Make a deep copy of the return from read_config(), since we modify
    # it during the profile name search.
    config = ConfigParser()
    config.read_dict(read_config())

    # (For 1.) Save the default profile and return it if its default farm matches.
    default_aws_profile_name: str = str(get_setting("defaults.aws_profile_name", config=config))
    if get_setting("defaults.farm_id", config=config) == farm_id:
        return default_aws_profile_name

    # (For 3.) We'll accumulate the profiles whose farms matched here
    first_farm_aws_profile_name: Optional[str] = None

    # (For 2.) Search for a profile with both a farm and queue match
    for aws_profile_name in aws_profile_names:
        set_setting("defaults.aws_profile_name", aws_profile_name, config=config)
        if get_setting("defaults.farm_id", config=config) == farm_id:
            # Return if both matched
            if queue_id and get_setting("defaults.queue_id", config=config) == queue_id:
                return aws_profile_name
            # Save if this was the first time the farm matched
            if not first_farm_aws_profile_name:
                first_farm_aws_profile_name = aws_profile_name

    # Return the first farm-matched profile, or the default if there was none
    return first_farm_aws_profile_name or default_aws_profile_name


def str2bool(value: str) -> bool:
    """
    Converts a string to boolean, accepting a variety of on/off,
    true/false, 0/1 variants.
    """
    value = value.lower()
    if value in _BOOL_VALUES:
        return value in _TRUE_VALUES
    else:
        raise ValueError(f"{value!r} is not a valid boolean string value")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Module for accessing the local file hash cache.
"""

import logging
from dataclasses import dataclass
from typing import Any, Dict, Optional

from .cache_db import CacheDB
from ..asset_manifests.hash_algorithms import HashAlgorithm


logger = logging.getLogger("Deadline")


@dataclass
class HashCacheEntry:
    """Represents an entry in the local hash-cache database"""

    # The file_path is stored as a BLOB in sqlite, encoded with utf-8 and the "surrogatepass"
    # error handler, as file names encountered in practice require this.
    file_path: str
    hash_algorithm: HashAlgorithm
    file_hash: str
    last_modified_time: str

    def to_dict(self) -> Dict[str, Any]:
        return {
            "file_path": self.file_path,
            "hash_algorithm": self.hash_algorithm.value,
            "file_hash": self.file_hash,
            "last_modified_time": self.last_modified_time,
        }


class HashCache(CacheDB):
    """
    Class used to store and retrieve entries in the local file hash cache.

    This class is intended to always be used with a context manager to properly
    close the connection to the hash cache database.

    This class also automatically locks when doing writes, so it can be called
    by multiple threads.
    """

    CACHE_NAME = "hash_cache"
    CACHE_DB_VERSION = 3

    def __init__(self, cache_dir: Optional[str] = None) -> None:
        table_name: str = f"hashesV{self.CACHE_DB_VERSION}"
        create_query: str = f"CREATE TABLE hashesV{self.CACHE_DB_VERSION}(file_path blob primary key, hash_algorithm text secondary key, file_hash text, last_modified_time timestamp)"
        super().__init__(
            cache_name=self.CACHE_NAME,
            table_name=table_name,
            create_query=create_query,
            cache_dir=cache_dir,
        )

    def get_entry(
        self, file_path_key: str, hash_algorithm: HashAlgorithm
    ) -> Optional[HashCacheEntry]:
        """
        Returns an entry from the hash cache, if it exists.
        """
        if not self.enabled:
            return None

        with self.db_lock, self.db_connection:
            entry_vals = self.db_connection.execute(
                f"SELECT * FROM {self.table_name} WHERE file_path=? AND hash_algorithm=?",
                [
                    file_path_key.encode(encoding="utf-8", errors="surrogatepass"),
                    hash_algorithm.value,
                ],
            ).fetchone()
            if entry_vals:
                return HashCacheEntry(
                    file_path=str(entry_vals[0], encoding="utf-8", errors="surrogatepass"),
                    hash_algorithm=HashAlgorithm(entry_vals[1]),
                    file_hash=entry_vals[2],
                    last_modified_time=str(entry_vals[3]),
                )
            else:
                return None

    def put_entry(self, entry: HashCacheEntry) -> None:
        """Inserts or replaces an entry into the hash cache database after acquiring the lock."""
        if self.enabled:
            with self.db_lock, self.db_connection:
                entry_dict = entry.to_dict()
                entry_dict["file_path"] = entry_dict["file_path"].encode(
                    encoding="utf-8", errors="surrogatepass"
                )
                self.db_connection.execute(
                    f"INSERT OR REPLACE INTO {self.table_name} VALUES(:file_path, :hash_algorithm, :file_hash, :last_modified_time)",
                    entry_dict,
                )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""Functions for interfacing with Deadline API calls."""

from typing import Optional

import boto3
from botocore.exceptions import ClientError

from ..exceptions import JobAttachmentsError
from ..models import (
    Attachments,
    JobAttachmentsFileSystem,
    Job,
    JobAttachmentS3Settings,
    ManifestProperties,
    PathFormat,
    Queue,
)
from .aws_clients import get_deadline_client


def get_queue(
    farm_id: str,
    queue_id: str,
    session: Optional[boto3.Session] = None,
    deadline_endpoint_url: Optional[str] = None,
) -> Queue:
    """
    Retrieves a specific queue from AWS Deadline Cloud.
    """
    try:
        response = get_deadline_client(
            session=session, endpoint_url=deadline_endpoint_url
        ).get_queue(farmId=farm_id, queueId=queue_id)
    except ClientError as exc:
        raise JobAttachmentsError(f'Failed to get queue "{queue_id}" from Deadline: {exc}') from exc

    # The API returns empty fields instead of an empty dict if there are no job attachment settings. So we need to
    # double check if the s3BucketName is set.
    if response.get("jobAttachmentSettings") and response["jobAttachmentSettings"].get(
        "s3BucketName"
    ):
        job_attachment_settings = JobAttachmentS3Settings(
            s3BucketName=response["jobAttachmentSettings"].get("s3BucketName", ""),
            rootPrefix=response["jobAttachmentSettings"].get("rootPrefix", ""),
        )
    else:
        job_attachment_settings = None

    display_name_key = "displayName"
    status_key = "status"
    if "name" in response:
        display_name_key = "name"
    if "state" in response:
        status_key = "state"

    return Queue(
        displayName=response[display_name_key],
        queueId=response["queueId"],
        farmId=response["farmId"],
        status=response[status_key],
        defaultBudgetAction=response["defaultBudgetAction"],
        jobAttachmentSettings=job_attachment_settings,
    )


def get_job(
    farm_id: str,
    queue_id: str,
    job_id: str,
    session: Optional[boto3.Session] = None,
    deadline_endpoint_url: Optional[str] = None,
) -> Job:
    """
    Retrieves a specific job from AWS Deadline Cloud.
    """
    try:
        response = get_deadline_client(session=session, endpoint_url=deadline_endpoint_url).get_job(
            farmId=farm_id, queueId=queue_id, jobId=job_id
        )
    except ClientError as exc:
        raise JobAttachmentsError(f'Failed to get job "{job_id}" from Deadline') from exc
    return Job(
        jobId=response["jobId"],
        attachments=(
            Attachments(
                manifests=[
                    ManifestProperties(
                        fileSystemLocationName=manifest_properties.get(
                            "fileSystemLocationName", None
                        ),
                        rootPath=manifest_properties["rootPath"],
                        rootPathFormat=PathFormat(manifest_properties["rootPathFormat"]),
                        outputRelativeDirectories=manifest_properties.get(
                            "outputRelativeDirectories", None
                        ),
                        inputManifestPath=manifest_properties.get("inputManifestPath", None),
                    )
                    for manifest_properties in response["attachments"]["manifests"]
                ],
                fileSystem=JobAttachmentsFileSystem(
                    response["attachments"].get("fileSystem", JobAttachmentsFileSystem.COPIED.value)
                ),
            )
            if "attachments" in response and response["attachments"]
            else None
        ),
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""Module that defines the second iteration of the asset manifest"""

from __future__ import annotations

import dataclasses
import json

from .base_manifest import BaseAssetManifest, BaseManifestPath


def canonical_path_comparator(path: BaseManifestPath):
    """
    Comparator for sorting paths.
    """
    # Sort by UTF-16 values as per the spec
    # https://www.rfc-editor.org/rfc/rfc8785.html#name-sorting-of-object-propertie
    # Use the "surrogatepass" error handler because filenames encountered in the wild
    # include surrogates.
    return path.path.encode("utf-16_be", errors="surrogatepass")


def manifest_to_canonical_json_string(manifest: BaseAssetManifest) -> str:
    """
    Return a canonicalized JSON string based on the following:
    * The JSON file *MUST* adhere to the JSON canonicalization guidelines
        outlined here (https://www.rfc-editor.org/rfc/rfc8785.html).
        * For now this is a simplification of this spec. Whitespace between JSON tokens are
            not emitted, and the keys are lexographically sorted. However the current implementation doesn't
            serialize Literals, String, Numbers, etc. to the letter of the spec explicitly.
            It implicitly follows the spec as the object keys all fall within the ASCII range of characters
            and this version of the Asset Manifest only serializes strings and integers.
    * The paths array *MUST* be in lexicographical order by path.
    """
    return json.dumps(
        dataclasses.asdict(manifest), sort_keys=True, separators=(",", ":"), ensure_ascii=True
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Functions for monitoring job status and retrieving job information.
"""

from __future__ import annotations
import datetime
import time
from typing import List, Dict, Any, Optional, Callable
from configparser import ConfigParser
from dataclasses import dataclass

from botocore.exceptions import ClientError

from deadline.client.exceptions import DeadlineOperationError, DeadlineOperationTimedOut
from deadline.client.api._session import (
    get_boto3_client,
    get_queue_user_boto3_session,
    get_user_and_identity_store_id,
)


@dataclass
class FailedTask:
    """
    Represents a failed task in a job.
    """

    step_id: str
    task_id: str
    step_name: str
    parameters: Dict[str, Any]
    session_id: Optional[str] = None


@dataclass
class JobCompletionResult:
    """
    Result of waiting for a job to complete.
    """

    status: str
    failed_tasks: List[FailedTask]
    elapsed_time: float


@dataclass
class LogEvent:
    """
    Represents a single log event from CloudWatch Logs.
    """

    timestamp: datetime.datetime
    message: str
    ingestion_time: Optional[datetime.datetime] = None
    event_id: Optional[str] = None


@dataclass
class SessionLogResult:
    """
    Result of retrieving logs for a session.
    """

    events: List[LogEvent]
    next_token: Optional[str]
    log_group: str
    log_stream: str
    count: int


def wait_for_job_completion(
    farm_id: str,
    queue_id: str,
    job_id: str,
    max_poll_interval: int = 120,
    timeout: int = 0,
    config: Optional[ConfigParser] = None,
    status_callback: Optional[Callable] = None,
) -> JobCompletionResult:
    """
    Wait for a job to complete and return information about its status and any failed tasks.

    This function blocks until the job's taskRunStatus reaches a terminal state
    (SUCCEEDED, FAILED, CANCELED, SUSPENDED, or NOT_COMPATIBLE), then returns a JobCompletionResult
    object containing the final status and any failed tasks.

    The function uses exponential backoff for polling, starting at 0.5 seconds and doubling
    the interval after each check until it reaches the maximum polling interval.

    Args:
        farm_id: The ID of the farm containing the job.
        queue_id: The ID of the queue containing the job.
        job_id: The ID of the job to wait for.
        max_poll_interval: Maximum time in seconds between status checks (default: 120).
        timeout: Maximum time in seconds to wait (0 for no timeout).
        config: Optional configuration object.
        status_callback: Optional callback function that receives the current status during polling.

    Returns:
        A JobCompletionResult object containing the job's final status and any failed tasks.

    Raises:
        DeadlineOperationError: If the timeout is reached or there's an error retrieving job information.
    """
    deadline = get_boto3_client("deadline", config=config)

    start_time = datetime.datetime.now()
    terminal_states = ["SUCCEEDED", "FAILED", "CANCELED", "SUSPENDED", "NOT_COMPATIBLE"]
    status = ""

    # Initial polling interval of 0.5 seconds
    current_interval = 0.5

    while True:
        # Check timeout
        if timeout > 0:
            elapsed = (datetime.datetime.now() - start_time).total_seconds()
            if elapsed > timeout:
                raise DeadlineOperationTimedOut(
                    f"Timeout waiting for job {job_id} to complete after {elapsed:.1f} seconds"
                )

        # Get job status
        try:
            job = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)
            status = job.get("taskRunStatus", "")

            # Call the status callback if provided with elapsed time and timeout info
            if status_callback:
                elapsed = (datetime.datetime.now() - start_time).total_seconds()
                status_callback(status, elapsed, timeout)

        except ClientError as exc:
            raise DeadlineOperationError(f"Failed to get job status: {exc}") from exc

        if status in terminal_states:
            break

        # Sleep using current interval
        time.sleep(current_interval)

        # Exponential backoff with a maximum interval
        current_interval = min(current_interval * 2, max_poll_interval)

    elapsed_time = (datetime.datetime.now() - start_time).total_seconds()

    # If job failed, collect failed tasks
    failed_tasks = []
    if status != "SUCCEEDED":
        try:
            # Get all steps with pagination
            paginator = deadline.get_paginator("list_steps")
            for page in paginator.paginate(farmId=farm_id, queueId=queue_id, jobId=job_id):
                # For each step, get tasks and filter for failed ones client-side
                for step in page["steps"]:
                    step_id = step["stepId"]
                    step_name = step.get("name", "")

                    # Only query for tasks if the step has any failed tasks
                    if step.get("taskRunStatusCounts", {}).get("FAILED", 0) > 0:
                        # Get all tasks with pagination
                        task_paginator = deadline.get_paginator("list_tasks")
                        for tasks_page in task_paginator.paginate(
                            farmId=farm_id, queueId=queue_id, jobId=job_id, stepId=step_id
                        ):
                            # Filter failed tasks client-side
                            for task in tasks_page["tasks"]:
                                if task.get("runStatus") == "FAILED":
                                    # Extract session ID from latestSessionActionId
                                    session_id = None
                                    latest_session_action_id = task.get("latestSessionActionId")
                                    if latest_session_action_id:
                                        # Format is typically "sessionaction-{session_id}-{action_number}"
                                        # Extract the session ID part
                                        parts = latest_session_action_id.split("-")
                                        if len(parts) >= 3 and parts[0] == "sessionaction":
                                            session_id = f"session-{parts[1]}"

                                    failed_tasks.append(
                                        FailedTask(
                                            step_id=step_id,
                                            task_id=task["taskId"],
                                            step_name=step_name,
                                            parameters=task.get("parameters", {}),
                                            session_id=session_id,
                                        )
                                    )
        except ClientError as exc:
            raise DeadlineOperationError(f"Failed to retrieve failed tasks: {exc}") from exc

    return JobCompletionResult(status=status, failed_tasks=failed_tasks, elapsed_time=elapsed_time)


def get_session_logs(
    farm_id: str,
    queue_id: str,
    session_id: str,
    limit: int = 100,
    start_time: Optional[datetime.datetime] = None,
    end_time: Optional[datetime.datetime] = None,
    next_token: Optional[str] = None,
    config: Optional[ConfigParser] = None,
) -> SessionLogResult:
    """
    Get CloudWatch logs for a specific session.

    This function retrieves logs from CloudWatch for the specified session ID.
    By default, it returns the most recent 100 log lines, but this can be
    adjusted using the limit parameter.

    Args:
        farm_id: The ID of the farm containing the session.
        queue_id: The ID of the queue containing the session.
        session_id: The ID of the session to get logs for.
        limit: Maximum number of log lines to return.
        start_time: Optional start time for logs as a datetime object.
        end_time: Optional end time for logs as a datetime object.
        next_token: Optional token for pagination of results.
        config: Optional configuration object.

    Returns:
        A SessionLogResult object containing the log events and metadata.

    Raises:
        DeadlineOperationError: If there's an error retrieving the logs.
    """
    # Get the Deadline client to use for getting queue credentials
    deadline = get_boto3_client("deadline", config=config)

    # Check if we have user and identity store ID (from Deadline Cloud monitor)
    user_id, identity_store_id = get_user_and_identity_store_id(config=config)

    # Create logs client - either with queue credentials or directly
    if user_id and identity_store_id:
        # Get a session with queue user credentials
        try:
            queue_session = get_queue_user_boto3_session(
                deadline=deadline, config=config, farm_id=farm_id, queue_id=queue_id
            )
            logs_client = queue_session.client("logs")
        except Exception as e:
            raise DeadlineOperationError(f"Failed to get queue credentials: {e}")
    else:
        # Use the same boto session as for deadline
        logs_client = get_boto3_client("logs", config=config)

    # Construct the log group name
    log_group_name = f"/aws/deadline/{farm_id}/{queue_id}"

    # Prepare parameters for GetLogEvents
    params = {
        "logGroupName": log_group_name,
        "logStreamName": session_id,
        "limit": limit,
        "startFromHead": False,  # Get the most recent logs first
    }

    # Add next_token if provided
    if next_token:
        params["nextToken"] = next_token

    # Add optional time parameters if provided
    if start_time:
        try:
            # Convert datetime to milliseconds since epoch
            start_timestamp = int(start_time.timestamp() * 1000)
            params["startTime"] = start_timestamp
        except (ValueError, AttributeError) as e:
            raise DeadlineOperationError(f"Invalid start time: {e}")

    if end_time:
        try:
            # Convert datetime to milliseconds since epoch
            end_timestamp = int(end_time.timestamp() * 1000)
            params["endTime"] = end_timestamp
        except (ValueError, AttributeError) as e:
            raise DeadlineOperationError(f"Invalid end time: {e}")

    try:
        response = logs_client.get_log_events(**params)

        # Convert to strongly typed objects
        events = []
        for event in response.get("events", []):
            events.append(
                LogEvent(
                    timestamp=datetime.datetime.fromtimestamp(
                        event["timestamp"] / 1000, tz=datetime.timezone.utc
                    ),
                    message=event["message"].rstrip(),
                    ingestion_time=(
                        datetime.datetime.fromtimestamp(event["ingestionTime"] / 1000)
                        if "ingestionTime" in event
                        else None
                    ),
                    event_id=event.get("eventId"),
                )
            )

        return SessionLogResult(
            events=events,
            next_token=response.get("nextForwardToken"),
            log_group=log_group_name,
            log_stream=session_id,
            count=len(events),
        )

    except logs_client.exceptions.ResourceNotFoundException:
        # Return an empty result if the log group or stream doesn't exist
        return SessionLogResult(
            events=[],
            next_token=None,
            log_group=log_group_name,
            log_stream=session_id,
            count=0,
        )
    except Exception as e:
        raise DeadlineOperationError(f"Failed to retrieve logs: {e}")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
UI widgets for the Scene Settings tab.
"""

import os

from qtpy.QtCore import Qt  # type: ignore
from qtpy.QtWidgets import (  # type: ignore
    QCheckBox,
    QComboBox,
    QGridLayout,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QTextEdit,
    QWidget,
)

from ..dataclasses import CliJobSettings
from .path_widgets import DirectoryPickerWidget


class CliJobSettingsWidget(QWidget):
    """
    Widget containing job setup specific to CLI jobs.

    Args:
        initial_settings (CliJobSettings): dataclass containing the job-specific settings.
        parent: The parent Qt Widget.
    """

    def __init__(self, initial_settings: CliJobSettings, parent=None):
        super().__init__(parent=parent)

        self._build_ui()
        self._load_initial_settings(initial_settings)

    def _set_enabled_with_label(self, prop_name: str, enabled: bool):
        """Set the enabled status of a control and its label"""
        getattr(self, prop_name).setEnabled(enabled)
        getattr(self, prop_name + "_label").setEnabled(enabled)

    def _build_ui(self):
        layout = QGridLayout(self)

        self.bash_script = QTextEdit()
        if os.name == "nt":
            font_family = "Consolas"
        elif os.name == "darwin":
            font_family = "Monaco"
        else:
            font_family = "Monospace"
        font = self.bash_script.currentFont()
        font.setFamily(font_family)
        font.setFixedPitch(True)
        font.setKerning(False)
        font.setPointSize(font.pointSize() + 1)
        self.bash_script.setCurrentFont(font)
        self.bash_script.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        layout.addWidget(self.bash_script, 0, 0, 1, 2)

        self.use_array_parameter_chck = QCheckBox("Use array parameter", self)
        self.array_parameter_name = QLineEdit(self)
        layout.addWidget(self.use_array_parameter_chck, 1, 0)
        layout.addWidget(self.array_parameter_name, 1, 1)
        self.use_array_parameter_chck.stateChanged.connect(self.use_array_parameter_changed)

        self.array_parameter_values_label = QLabel("Array parameter values")
        layout.addWidget(self.array_parameter_values_label, 2, 0)
        self.array_parameter_values = QLineEdit(self)
        layout.addWidget(self.array_parameter_values, 2, 1)

        self.data_dir_label = QLabel("Data directory")
        self.data_dir_edit = DirectoryPickerWidget(
            initial_directory=os.path.expanduser(os.path.join("~", "CLIJobData")),
            directory_label="Data directory",
            parent=self,
        )
        layout.addWidget(self.data_dir_label, 3, 0)
        layout.addWidget(self.data_dir_edit, 3, 1)

        self.file_format_label = QLabel("Template file format")
        self.file_format_box = QComboBox(parent=self)
        self.file_format_box.addItems(["YAML", "JSON"])
        layout.addWidget(self.file_format_label, 4, 0)
        layout.addWidget(self.file_format_box, 4, 1)

    def _load_initial_settings(self, initial_settings: CliJobSettings):
        self.bash_script.setPlainText(initial_settings.bash_script_contents)
        self.use_array_parameter_chck.setChecked(initial_settings.use_array_parameter)
        self.array_parameter_name.setText(initial_settings.array_parameter_name)
        self.array_parameter_values.setText(initial_settings.array_parameter_values)
        self.file_format_box.setCurrentText(initial_settings.file_format)

    def update_settings(self, settings: CliJobSettings):
        """
        Update a settings object with the latest values.
        """
        settings.bash_script_contents = self.bash_script.toPlainText()
        settings.use_array_parameter = self.use_array_parameter_chck.isChecked()
        settings.array_parameter_name = self.array_parameter_name.text()
        settings.array_parameter_values = self.array_parameter_values.text()
        settings.data_dir = os.path.expanduser(self.data_dir_edit.text())
        settings.file_format = self.file_format_box.currentText()

    def use_array_parameter_changed(self, state: int):
        enabled = Qt.CheckState(state) == Qt.Checked
        self.array_parameter_name.setEnabled(enabled)
        self._set_enabled_with_label("array_parameter_values", enabled)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
import glob
import json
from pathlib import Path
from typing import List, Optional
from deadline.client.exceptions import NonValidInputError
from deadline.job_attachments.models import GlobConfig


def _process_glob_inputs(glob_arg_input: str) -> GlobConfig:
    """
    Helper function to process glob inputs.
    glob_input: String, can represent a json, filepath, or general include glob syntax.
    """

    # Default Glob config.
    glob_config = GlobConfig()
    if glob_arg_input is None or len(glob_arg_input) == 0:
        # Not configured, or not passed in.
        return glob_config

    try:
        input_as_path = Path(glob_arg_input)
        if input_as_path.is_file():
            # Read the file so it can be parsed as JSON.
            with open(glob_arg_input) as f:
                glob_arg_input = f.read()
    except Exception:
        # If this cannot be processed as a file, try it as JSON.
        pass

    try:
        # Parse the input as JSON, default to Glob Config defaults.
        input_as_json = json.loads(glob_arg_input)
        glob_config.include_glob = input_as_json.get(GlobConfig.INCLUDE, glob_config.include_glob)
        glob_config.exclude_glob = input_as_json.get(GlobConfig.EXCLUDE, glob_config.exclude_glob)
    except Exception:
        # This is not a JSON blob, bad input.
        raise NonValidInputError(f"Glob input {glob_arg_input} cannot be deserialized as JSON")

    return glob_config


def _match_files_with_pattern(base_path: str, patterns: List[str]) -> set:
    """
    Helper function to match files based on glob patterns.

    Args:
        base_path: Root path to glob from
        patterns: List of glob patterns to match

    Returns:
        Set of normalized file paths that match the patterns
    """
    matched_files = set()
    for pattern in patterns:
        # Make pattern relative to base path
        full_pattern = os.path.join(base_path, pattern)

        # Use recursive glob for directory matching
        for matched_path in glob.glob(full_pattern, recursive=True):
            # Only add files, not directories
            if os.path.isfile(matched_path):
                # Convert to proper path format
                normalized_path = os.path.normpath(matched_path)
                matched_files.add(normalized_path)

    return matched_files


def _glob_paths(
    path: str, include: List[str] = ["**/*"], exclude: Optional[List[str]] = None
) -> List[str]:
    """
    Glob routine that supports Unix style pathname pattern expansion for includes and excludes.
    This function will recursively list all files of path, including all files globbed by include and removing all files marked by exclude.
    path: Root path to glob.
    include: Optional, pattern syntax for files to include.
    exclude: Optional, pattern syntax for files to exclude.
    return: List of files found based on supplied glob patterns.
    """
    # Convert path to absolute path
    base_path = os.path.abspath(path)

    # Process include patterns
    matched_files = _match_files_with_pattern(base_path, include)

    # Process exclude patterns
    if exclude:
        files_to_exclude = _match_files_with_pattern(base_path, exclude)
        # Remove excluded files from result
        matched_files -= files_to_exclude

    return list(matched_files)
